"""
atoms
=======================

This module implements the generic ``Atom`` class, that can be used to define
a custom atom species.

Note
----
    In general, it is better to create new classes for specific
    atomic species, so that they be used by all users. Specifi atomic species
    are stored in ``atomsmltr.atoms.collection``

"""

# % IMPORTS
from abc import ABC, abstractmethod
from copy import copy

import scipy.constants as csts
import numpy as np

# % LOCAL IMPORTS
from .transitions import AtomicTransition
from ..utils.infostring import InfoString

# % ABSTRACT CLASSES


class Atom(ABC):
    """A generic class to define atomic species.

    Parameters
    ----------
    mass : float
        the atomic mass in kg
    name : str
        a name to describe the atom


    Example
    -------
    >>> from atomsmltr.atoms import Atom
    >>> from scipy import constants as csts
    >>> atom = Atom(mass=4 * csts.m_u, name="Helium")

    Note
    ----
    Specific atomic species are implemented in ``atomsmltr.atoms.collection``

    """

    def __init__(self, mass: float, name: str):
        self.__mass = mass
        self.__name = name
        self.__transitions = {}
        super().__init__()

    # - GETTERS AND SETTERS

    @property
    def mass(self) -> float:
        """float: the atom mass in kg"""
        return self.__mass

    @property
    def mass_au(self) -> float:
        """float: the atom mass in atomic units"""
        return self.__mass / csts.m_u

    @property
    def name(self) -> float:
        """str: the atom name"""
        return self.__name

    @property
    def transitions(self) -> list:
        """list: a list of the atom's transitions"""
        return self.__transitions.values()

    @property
    def trans(self) -> dict:
        """dict: a copy of the atom's transition dictionnary"""
        return copy(self.__transitions)

    # - RADIATION PRESSURE

    def get_radiation_pressure(
        self,
        transition: str,  # the transition tag
        intensity: float,  # the intensity in W/cm^2
        mag_field: float,  # the amplitude of the magnetic field
        polarization: np.ndarray,  # projection of laser polarization on (pi, sigma+, sigma-)
        detuning: float,  # laser detuning
    ) -> float:
        """Computes the radiation pressure.

        Parameters
        ----------
        transition : str
            The tag of the considered transition. The transition has to be defined in the atom's
            transition list, that can be accessed using the ``list_transitions()`` method.

        intensity: float
            The laser intensity, in W/m^2

        mag_field: float
            The *norm* of the magnetic field, in T (scalar).

        polarization: ndarray, shape (,3)
            Projection of the laser polarization on (pi, sigma+, sigma-)

        detuning: float
            The laser bare detuning, in rad/s

        Returns
        -------
        F_rad: float
            The radiation pressure in SI (scalar)

        Notes
        -------
        Yields the radiation pressure felt by an atom excited on a given transition, by
        a laser with given intensity. It takes into account the value of the magnetic field,
        the polarization of the laser and its bare detuning.

        The computation is based on the ``get_scattering_rate()`` method, implemented in the
        ``AtomicTransition`` class.

        See Also
        --------
        atomsmltr.atoms.transitions.AtomicTransition.get_scattering_rate()

        """
        # parameter check
        transition_list = self.list_transitions()
        msg = f"There is no transition with tag '{transition}'. Available transitions : {transition_list}"
        if transition not in transition_list:
            raise KeyError(msg)

        # get scattering rate
        trans = self.__transitions[transition]
        scattering_rate = trans.get_scattering_rate(
            intensity, mag_field, polarization, detuning
        )

        # convert to radiation pressure
        k = 2 * csts.pi / trans.wavelength
        F_rad = csts.hbar * k * scattering_rate

        return F_rad

    # - TRANSITIONS MANAGEMENT

    def get_transitions_copy(self) -> dict:
        """returns a copy of the transition dictionnary

        Returns
        -------
        dict
            a dict containing the atom's transitions
        """
        return copy(self.__transitions)

    def add_transition(self, transition: AtomicTransition, tag=None) -> None:
        """adds a transition to the atom transition dict.

        Parameters
        ----------
        transition : AtomicTransition
            an atomic transition (see ``atomsmltr.atoms.transitions.AtomicTransition``)
        tag : str, optional
            the tag identifying the transition in the transition dict.
            If nothing or ``None`` is given, we will use the ``.tag`` property
            of the ``AtomicTransition`` object

        See Also
        --------
        atomsmltr.atoms.transitions.AtomicTransition
        """
        # -- parse input
        # check transition type
        if not isinstance(transition, AtomicTransition):
            raise TypeError("`transition` should be an `AtomicTransition`object.")
        # if tag is non, use the atomic transition builtin tag
        tag = transition.tag if tag is None else tag
        # check that not in dictionnary
        msg = f"There is already a transition with the tag {tag} in the atom's collection."
        assert tag not in self.__transitions, msg
        # TODO: other checks ? warning if already same wavelength ?

        # -- add to collection
        self.__transitions.update({tag: transition})

    def list_transitions(self) -> list:
        """returns a list of the transition tags (str)

        Returns
        -------
        list of str
            list of transition tags
        """
        return list(self.__transitions.keys())

    def rm_transition(self, tag: str) -> None:
        """removes a transition for the atom's transition dict.

        Parameters
        ----------
        tag : str
            the tag of the transition to remove
        """
        del self.__transitions[tag]

    # - INFOSTRING

    def _gen_infostring_obj(self) -> InfoString:
        """generates an ``InfoString`` object.

        Returns
        -------
        InfoString
            an ``InfoString`` object

        See also
        --------
        atomsmltr.utils.infostring.InfoString
        """

        info = InfoString(title=self.name)
        info.add_section("Parameters")
        info.add_element("mass (kg)", f"{self.mass:.2e}")
        info.add_element("mass (au)", f"{self.mass_au:.2f}")

        info.add_section("Transition list")
        for trans_tag in self.list_transitions():
            info.add_element(trans_tag)

        for trans_tag, trans in self.__transitions.items():
            trans_info = trans.gen_infostring_obj()
            section_title = f"'{trans_tag}' transition"
            info.absorb_section(trans_info, "Parameters", section_title)

        return info

    def gen_infostring_obj(self) -> InfoString:
        """generates an ``InfoString`` object.

        Returns
        -------
        InfoString
            an ``InfoString`` object

        See also
        --------
        atomsmltr.utils.infostring.InfoString
        """
        return self._gen_infostring_obj()

    def gen_info_string(self, **kwargs) -> str:
        """generates an info string

        Returns
        -------
        info_string: str
            a string with information on the atom
        """
        return self.gen_infostring_obj().generate(**kwargs)

    def print_info(self) -> None:
        """prints the atom infostring"""
        print(self.gen_info_string())



"""
Examples: Feng 2024
=======================

This example provides the configuration for the Strontium source described
in Feng et al., 2024, Quantum Sci. Technol. 9, 025017 (DOI: 10.1088/2058-9565/ad310b).

This example requires Magpylib.
"""

# % EXPERIENCE DESCRIPTION
description = """

Strontium 2D MOT with Zeeman slower and symmetric permanent magnet configuration,
from Feng et al 2024 Quantum Sci. Technol. 9 025017 (DOI 10.1088/2058-9565/ad310b).

>> from atomsmltr.examples.feng2024 import config

This simulation models a cold Strontium atomic source, consisting of a combination of
permanent magnets, Zeeman slower lasers, and a two-dimensional magneto-optical trap (2D MOT).
The magnetic field is generated by four symmetric sets of permanent magnets, producing
a linear quadrupole magnetic field, which vanished along the y-axis. The Zeeman slower decelerates atoms along the z-axis,
while the 2D MOT provides cooling and confinement in the transverse directions.

The launch principle relies on red-detuned laser beams interacting with the atoms: counter-propagating
lasers in the Zeeman slower slow atoms along z, and orthogonal 2D MOT beams create trapping forces
in the x-z and y-z planes.

Physical configuration:
- Atom species : 88Sr (main transition at 461 nm)
- Magnetic field : symmetric configuration of permanent cuboid magnets
  - Magnetization : -8.7e5 A/m along x
  - Dimensions : 3 × 10 × 25 mm
  - Arrangement : 4 cubes, each with 9 magnets along x, rotated ±180° along the y-axis for symmetry
- Zeeman slower lasers 1:
  - Wavelength : 461 nm
  - Waist : 6.7 mm
  - Power : 45 mW
  - Polarization : σ+/σ−
  - Direction : along negative z-axis
  - Detuning : -13 Γ
- Zeeman slower lasers 2:
  -Wavelength : 461 nm
  - Waist : 7.2 mm
  - Power : 85 mW
  - Polarization : σ+/σ−
  - Direction : along negative z-axis
  - Detuning : -13 Γ
- 2D MOT lasers:
  - Wavelength : 461 nm
  - Waist : 12 mm
  - Power : 80 mW per beam
  - Polarization : σ+/σ−
  - Directions : tilted diagonally in x-z and y-z planes (±45°)
  - Detuning : -1.15 Γ

"""

# % IMPORTS

import magpylib as magpy
import numpy as np

# % LOCAL IMPORTS
from ..atoms import Strontium
from ..environment.lasers import GaussianLaserBeam
from ..environment.lasers.polarization import CircularLeft, CircularRight, Horizontal
from ..environment.fields.magnetic.magpylib import MagpylibWrapper
from ..environment.zones import Limits
from ..simulation import Configuration

# --------------------------------------------------------------------------------------------------------------------------------

# % GENERATE ELEMENTS OF THE CONFIGURATIONS

# -- setup lasers
# cf. config from Feng et al. 2024
l461_ZS = {
    "wavelength": 461e-9,
    "waist": 6.7e-3,
    "power": 90e-3,
    "waist_position": (0, 0, 0),
}
l461_ZS2 = {
    "wavelength": 461e-9,
    "waist": 7.2e-3,
    "power": 150e-3,
    "waist_position": (0, 0, 0),
}
l461_2DMOT = {
    "wavelength": 461e-9,
    "waist": 12e-3,
    "power": 80e-3,
    "waist_position": (0, 0, 0),
}

# -- setup atom
atom_strontium = Strontium()

# -- get Strontium main transition information
main = atom_strontium.trans["main"]

# --------------------------------------------------------------------------------------------------------------------------------

# GENERATE SYMMETRIC FIELD

# -- setup magnetic field: symmetric configuration of permanent magnets in Feng et a 2024
# Define magnet properties
magnetization = (-8.7e5, 0, 0)
dimension = (0.003, 0.010, 0.025)

# X positions for the cuboids (same as in your original code)
x_positions = np.linspace(-0.012, 0.012, 9)

# Create cube1 and cube2 for symetric field (lists of magnets)
cube1 = [
    magpy.magnet.Cuboid(magnetization=magnetization, dimension=dimension)
    for _ in x_positions
]
cube2 = [
    magpy.magnet.Cuboid(magnetization=magnetization, dimension=dimension)
    for _ in x_positions
]

# Assign positions to cube1 and cube2
for i, x in enumerate(x_positions):
    cube1[i].position = (x, 0.039, -0.050)
    cube2[i].position = (x, -0.039, -0.050)

# Create cube3 and cube4 as copies of cube1 and cube2
cube3 = [c.copy() for c in cube1]
cube4 = [c.copy() for c in cube2]

# Rotate cube3 and cube4 by 180 degrees along the y-axis
for c in cube3 + cube4:
    c.rotate_from_angax(180, "y", anchor=0)

# Combine all cubes into magnetSet as a single MagnetSet
magnetSet = magpy.Collection(cube1 + cube2 + cube3 + cube4)

# wrap it up
mag_field = MagpylibWrapper(magnetSet)
mag_field.tag = "Symmetric Permanent Magnet Configuration"


# --------------------------------------------------------------------------------------------------------------------------------

# % GENERATE ASYMMETRIC FIELD

# -- setup magnetic field: asymmetric configuration of permanent magnets in Feng et a 2024
# Define magnet properties
magnetization_asymmetric = (-8.7e5, 0, 0)
inv_magnetization = (8.7e5, 0, 0)
dimension_asymmetric = (0.003, 0.010, 0.025)


# X positions for the cuboids (same as in your original code)
x_positions_1_2 = np.linspace(-0.012, 0.012, 9)  # Positions for cube1 and cube2
x_positions_3_4 = np.linspace(-0.018, 0.018, 13)  # Positions for cube3 and cube4
x_positions_1_2 = np.linspace(-0.012, 0.012, 9)  # Positions for cube1 and cube2
x_positions_3_4 = np.linspace(-0.018, 0.018, 13)  # Positions for cube3 and cube4

# Create cube1 and cube2 (lists of magnets)
cube1_asymmetric = [
    magpy.magnet.Cuboid(
        magnetization=magnetization_asymmetric, dimension=dimension_asymmetric
    )
    for _ in x_positions_1_2
]
cube2_asymmetric = [
    magpy.magnet.Cuboid(
        magnetization=magnetization_asymmetric, dimension=dimension_asymmetric
    )
    for _ in x_positions_1_2
]

# Assign positions to cube1 and cube2
for i, x in enumerate(x_positions_1_2):
    cube1_asymmetric[i].position = (x, 0.039, -0.050)
    cube2_asymmetric[i].position = (x, -0.039, -0.050)

# Create cube3 and cube4 with 13 elements each
cube3_asymmetric = [
    magpy.magnet.Cuboid(magnetization=inv_magnetization, dimension=dimension_asymmetric)
    for _ in x_positions_3_4
]
cube4_asymmetric = [
    magpy.magnet.Cuboid(magnetization=inv_magnetization, dimension=dimension_asymmetric)
    for _ in x_positions_3_4
]

# Assign positions to cube3 and cube4
for i, x in enumerate(x_positions_3_4):
    cube3_asymmetric[i].position = (x, 0.035, 0.0623)
    cube4_asymmetric[i].position = (x, -0.035, 0.0623)

# Combine all cubes into magnetSet as a single MagnetSet
magnetSet = magpy.Collection(
    cube1_asymmetric + cube2_asymmetric + cube3_asymmetric + cube4_asymmetric
)

# wrap it up
mag_field_asymmetric = MagpylibWrapper(magnetSet)
mag_field_asymmetric.tag = "Asymmetric Permanent Magnet Configuration"

# --------------------------------------------------------------------------------------------------------------------------------


# UPDATED FUBCTION TO ALLOW SPECIFIC DIRECTION
def create_laser(tag, lasers_dict, params, direction, polarization=None):
    laser = GaussianLaserBeam(**params)
    laser.direction = np.array(direction)
    laser.tag = tag
    if polarization is not None:
        laser.polarization = polarization
    lasers_dict[tag] = laser


# --------------------------------------------------------------------------------------------------------------------------------

# % GENERATE CONFIGURATION WITH SYMMETRIC MAGNETIC FIELD

# -- init config with strontium atom
config_symmetric_field = Configuration()
config_symmetric_field.atom = atom_strontium

# -- init laser list for the symmetric config
lasers_symmetric = {}

# -- create Zeeman slower lasers
create_laser("ZS1", lasers_symmetric, l461_ZS, [0, 0, -1], Horizontal())

# -- define directions for MM 2D MOT
d1_dir = np.array([1, 1, 1]) / np.sqrt(3)
d2_dir = np.array([-1, 1, -1]) / np.sqrt(3)
d3_dir = np.array([1, -1, -1]) / np.sqrt(3)
d4_dir = np.array([-1, -1, 1]) / np.sqrt(3)

# -- create 2D MOT lasers
create_laser("d1<", lasers_symmetric, l461_2DMOT, d1_dir, CircularLeft())
create_laser("d1>", lasers_symmetric, l461_2DMOT, d2_dir, CircularLeft())
create_laser("d2<", lasers_symmetric, l461_2DMOT, d3_dir, CircularRight())
create_laser("d2>", lasers_symmetric, l461_2DMOT, d4_dir, CircularRight())

# -- configure the zone limits
xlim_symmetric = Limits(
    -0.15, 0.35, axis=2, target="position", action="stop", tag="xlim"
)

# -- add objects
config_symmetric_field += (
    xlim_symmetric,
    lasers_symmetric["d1>"],
    lasers_symmetric["d1<"],
    lasers_symmetric["d2>"],
    lasers_symmetric["d2<"],
    lasers_symmetric["ZS1"],
    mag_field,
)

# -- setup atomlight
for laser in config_symmetric_field.list_lasers():
    if laser == "ZS1":
        config_symmetric_field.add_atomlight_coupling(
            laser=laser, transition="main", detuning=-13 * main.Gamma
        )
    if (laser == "d1>") or (laser == "d1<"):
        config_symmetric_field.add_atomlight_coupling(
            laser=laser, transition="main", detuning=-1.0 * main.Gamma
        )
    if (laser == "d2>") or (laser == "d2<"):
        config_symmetric_field.add_atomlight_coupling(
            laser=laser, transition="main", detuning=-1.4 * main.Gamma
        )


# --------------------------------------------------------------------------------------------------------------------------------

# % GENERATE CONFIGURATION WITH ASYMMETRIC MAGNETIC FIELD AND A SINGLE FREQUENCY LASER

# -- init config with strontium atom
config_asymmetric_field_1 = Configuration()
config_asymmetric_field_1.atom = atom_strontium

# -- init laser list for the symmetric config
lasers_asymmetric_1 = {}

# -- create Zeeman slower lasers
create_laser("ZS1", lasers_asymmetric_1, l461_ZS, [0, 0, -1], Horizontal())

# -- define directions for 2D MOT
d1_dir = np.array([1, 0, 1]) / np.sqrt(2)
d2_dir = np.array([1, 0, -1]) / np.sqrt(2)

# -- create 2D MOT lasers
create_laser("d1<", lasers_asymmetric_1, l461_2DMOT, -d1_dir, CircularLeft())
create_laser("d1>", lasers_asymmetric_1, l461_2DMOT, d1_dir, CircularLeft())
create_laser("d2<", lasers_asymmetric_1, l461_2DMOT, -d2_dir, CircularRight())
create_laser("d2>", lasers_asymmetric_1, l461_2DMOT, d2_dir, CircularRight())

# -- add everything to the configuration
xlim_asymmetric_1 = Limits(
    -0.15, 0.35, axis=0, target="position", action="stop", tag="xlim"
)

# -- add objects
config_asymmetric_field_1 += (
    xlim_asymmetric_1,
    lasers_asymmetric_1["d1>"],
    lasers_asymmetric_1["d1<"],
    lasers_asymmetric_1["d2>"],
    lasers_asymmetric_1["d2<"],
    lasers_asymmetric_1["ZS1"],
    mag_field_asymmetric,
)

# -- setup atomlight
for laser in config_asymmetric_field_1.list_lasers():
    if laser == "ZS1":
        config_asymmetric_field_1.add_atomlight_coupling(
            laser=laser, transition="main", detuning=-14 * main.Gamma
        )
    if (laser == "d1>") or (laser == "d1<") or (laser == "d2>") or (laser == "d2<"):
        config_asymmetric_field_1.add_atomlight_coupling(
            laser=laser, transition="main", detuning=-1.15 * main.Gamma
        )


# --------------------------------------------------------------------------------------------------------------------------------

# % GENERATE CONFIGURATION WITH ASYMMETRIC MAGNETIC FIELD AND A TWO FREQUENCY LASER

# -- init config with strontium atom
config_asymmetric_field_2 = Configuration()
config_asymmetric_field_2.atom = atom_strontium

# -- init laser list for the symmetric config
lasers_asymmetric_2 = {}

# -- create Zeeman slower lasers
create_laser("ZS1", lasers_asymmetric_2, l461_ZS, [0, 0, -1], Horizontal())
create_laser("ZS2", lasers_asymmetric_2, l461_ZS, [0, 0, -1], Horizontal())


# -- create 2D MOT lasers
create_laser("d1<", lasers_asymmetric_2, l461_2DMOT, d1_dir, CircularLeft())
create_laser("d1>", lasers_asymmetric_2, l461_2DMOT, d2_dir, CircularLeft())
create_laser("d2<", lasers_asymmetric_2, l461_2DMOT, d3_dir, CircularRight())
create_laser("d2>", lasers_asymmetric_2, l461_2DMOT, d4_dir, CircularRight())

# -- add everything to the configuration
xlim_asymmetric_2 = Limits(
    -0.15, 0.35, axis=0, target="position", action="stop", tag="xlim"
)

# -- add objects
config_asymmetric_field_2 += (
    xlim_asymmetric_1,
    lasers_asymmetric_2["d1>"],
    lasers_asymmetric_2["d1<"],
    lasers_asymmetric_2["d2>"],
    lasers_asymmetric_2["d2<"],
    lasers_asymmetric_2["ZS1"],
    lasers_asymmetric_2["ZS2"],
    mag_field_asymmetric,
)

# -- setup atomlight
for laser in config_asymmetric_field_2.list_lasers():
    if laser == "ZS1":
        config_asymmetric_field_2.add_atomlight_coupling(
            laser=laser, transition="main", detuning=-13 * main.Gamma
        )
    if (laser == "d1>") or (laser == "d1<"):
        config_asymmetric_field_2.add_atomlight_coupling(
            laser=laser, transition="main", detuning=-1.0 * main.Gamma
        )
    if (laser == "d2>") or (laser == "d2<"):
        config_asymmetric_field_2.add_atomlight_coupling(
            laser=laser, transition="main", detuning=-1.4 * main.Gamma
        )
    if laser == "ZS2":
        config_asymmetric_field_2.add_atomlight_coupling(
            laser=laser, transition="main", detuning=-28.5 * main.Gamma
        )



"""misc
==================

a collection of useful functions

"""

# % IMPORTS
import numpy as np
from random import choice

# % ARGUMENT PROCESSORS / CHECKERS


def check_positive_float(param_name: str, value: float) -> None:
    """internal function to check that a parameter is a positive float, raises a `ValueError` if not.

    Args:
        param_name (str): name of the checked parameter, to give context in the exception
        value (float): value of the paramater to check
    """
    if isinstance(value, int):
        value = float(value)
    if not isinstance(value, float):
        raise ValueError(f"'{param_name}' has to be a float")
    if value < 0:
        raise ValueError(f"'{param_name}' has to be a positive")


def check_position_array(position, nocheck=False):
    """Checks that a position array matches our vectorization convention.

    It raises an error if the shape is not good.

    Parameters
    ----------
    position : array
        the array to check
    nocheck : bool, optional
        if set to True, the function is bypasses

    Returns
    -------
    position
        the array

    Notes
    ------
    positions array should have a shape (3,) or (n1, n2, .., 3).

    In all cases, the last dimension contains cordinates (x, y, z),
    in meter and in the lab frame
    """
    if nocheck:
        return position
    # convert to array
    position = np.asanyarray(position)
    # check that shape is fine : should be (3,) or (n,3)
    if not position.shape or position.shape[-1] != 3:
        raise ValueError("The position array should be of shape (3,) or (n, m, .., 3)")
    return position


def check_position_speed_array(position, nocheck=False):
    """Checks that a position array matches our vectorization convention.

    It raises an error if the shape is not good.

    Parameters
    ----------
    position : array
        the array to check
    nocheck : bool, optional
        if set to True, the function is bypasses

    Returns
    -------
    position
        the array

    Notes
    ------
    positions array should have a shape (6,) or (n1, n2, .., 6).

    In all cases, the last dimension contains cordinates (x, y, z, vx, vy, vz),
    in meter and in the lab frame
    """
    if nocheck:
        return position
    # convert to array
    position = np.asanyarray(position)
    # check that shape is fine : should be (6,) or (n,6)
    if not position.shape or position.shape[-1] != 6:
        raise ValueError("The position array should be of shape (6,) or (n, m, .., 6)")
    return position


def check_scalar_field_value_function(func):
    """Used in tests : checks that a function yielding values of a 3D **scalar** field
    field behaves correctly with numpy arrays.

    for input of shape (..., 1) should return shape (..., 1)

    Parameters
    ----------
    func : function
        the function to check
    """

    # - 1 check that it works with a single position
    position = (0, 0, 0)
    value = func(position)
    assert value.ndim == 0

    # - 2 with arrays
    # -
    position = np.mgrid[0:1:8j, 0:5:10j, 0:0:1j].T
    X, _, _ = position.T
    X = X.T
    value = func(position)
    assert value.shape == X.shape
    # -
    position = position[0]
    X, _, _ = position.T
    X = X.T
    value = func(position)
    assert value.shape == X.shape


def check_vector_field_value_function(func):
    """Used in tests : checks that a function yielding values of a 3D **vector** field
    field behaves correctly with numpy arrays.

    for input of shape (..., 3) should return shape (..., 3)

    Parameters
    ----------
    func : function
        the function to check
    """

    # - 1 check that it works with a single position
    position = (0, 0, 0)
    value = func(position)
    assert value.shape == (3,)

    # - 2 with arrays
    # -
    position = np.mgrid[0:1:8j, 0:5:10j, 0:0:1j].T
    value = func(position)
    assert value.shape == position.shape
    # -
    position = position[0]
    value = func(position)
    assert value.shape == position.shape


# % RANDOM TAG


def random_word(syl=3):
    """Generates a random word

    Parameters
    ----------
    syl : int, optional
        number of syllabs, by default 3

    Returns
    -------
    word: str
        the random word
    """
    voy = "aeiou"
    cons = "zrtpqsdfghklmwxvbn"
    res = ""
    for i in range(syl):
        res += choice(cons) + choice(voy)
    return res



"""infostrings
==================

Here we implement the ``InfoString`` class, that allows to gather
different parameters from objects and to display them in a tree format.

This is used in the ``print_info()`` method that is present in most
classes accross the module.

Examples
--------------------

Run a simulation with one initial condition vector

.. code-block:: python

        from atomsmltr.utils.infostring import InfoString

        info = InfoString("PARAMETERS")
        info.add_section("First Section")
        info.add_element("elem", "param")
        info.add_element("x", f"{25.4e-6:.2e}µm")
        info.add_element("y", f"{15.8:.2f} mW")
        info.add_section("Another section")
        info.add_element("elem", "param")
        info.add_element("elem2", "param2")

        print(info.generate())

This prints :

.. code-block::

        ──────────────
        | PARAMETERS |
        ──────────────
        . First Section :
        ├── elem : param
        ├── x : 2.54e-05µm
        └── y : 15.80 mW

        . Another section :
        ├── elem : param
        └── elem2 : param2

"""

# % IMPORTS

from collections import OrderedDict

# % CONSTANTS

HEADER = ". {} :\n"
PARAM = "  ├── {} : {}\n"
PARAMSINGLE = "  ├── {}\n"
LPARAM = "  └── {} : {}\n\n"
LPARAMSINGLE = "  └── {}\n\n"
TITLE = "| {} |\n"

# % CLASS


class InfoString(object):
    """Generates info strings

    Parameters
    ----------
    title : str
        the title of the infostring object
    """

    def __init__(self, title: str):
        self.__title = title
        self.__elements = OrderedDict()
        self.__current_section = ""

    @property
    def elements(self):
        """OrderedDict: the elements inside the infostring object"""
        return self.__elements

    @property
    def title(self) -> str:
        """str: titile of the infostring object"""
        return self.title

    @title.setter
    def title(self, value: str):
        assert isinstance(value, str), "title should be a string"
        self.__title = value

    def add_section(self, name: str):
        """adds a new section for parameters

        Parameters
        ----------
        name : str
            the section name
        """
        if name in self.__elements:
            raise Warning(f"section '{name}' already exists")
        self.__elements[name] = OrderedDict()
        self.__current_section = name

    def add_element(self, name: str, value: str = None, section: str = None):
        """adds an element in a given section

        Parameters
        ----------
        name : str
            name of the element
        value : str, optional
            value of the element, by default None
        section : str, optional
            name of the section. If None is given, the element is added to
            the lastly used section, by default None
        """
        if section is None:
            section = self.__current_section
        if section not in self.__elements:
            raise Warning(f"section '{section}' does not exist")
        # switch current section
        self.__current_section = section
        if name in self.__elements[section]:
            raise Warning(f"section '{section}' already has an element {name}")
        self.__elements[section][name] = value

    def rm_element(self, name: str, section: str = None):
        """removes an element from a section

        Parameters
        ----------
        name : str
            name of the element to remove
        section : str, optional
            name of the section. If None is given, the element is removed from
            the lastly used section, by default None
        """

        if section is None:
            section = self.__current_section
        if section not in self.__elements:
            raise Warning(f"section '{section}' does not exist")
        # switch current section
        self.__current_section = section
        if name not in self.__elements[section]:
            raise Warning(f"section '{section}' does not have an element {name}")
        del self.__elements[section][name]

    def absorb_section(self, info, target_section: str, new_name: str = None):
        """incorporates a section from another infostring object

        Parameters
        ----------
        info : InfoString
            the infostring object from which we take the section
        target_section : str
            name of the section to incorporate
        new_name : str, optional
            name of the incorporated section.
            If None is given we use the name of the original section, by default None
        """
        # assert
        assert isinstance(info, InfoString), "'info' should be an InfoString object"
        # get info
        info_dic = info.elements
        assert (
            target_section in info_dic
        ), f"This info object does not have a '{target_section}' section"
        if new_name is None:
            new_name = target_section
        # absorb
        self.add_section(new_name)
        for name, value in info_dic[target_section].items():
            self.add_element(name, value)

    def merge(self, info, prefix=""):
        """merges an entire infostring

        Parameters
        ----------
        info : InfoString
            infostring object to merge
        prefix : str, optional
            prefix added to the names of the sections incorporated
            from the merged infostring, by default ""
        """
        # assert
        assert isinstance(info, InfoString), "'info' should be an InfoString object"
        # merge
        for section, elements in info.elements.items():
            self.add_section(prefix + section)
            for name, value in elements.items():
                self.add_element(name, value)

    def generate(self, display_title=True):
        """generates a string from the infostring object

        Parameters
        ----------
        display_title : bool, optional
            whether to display the title, by default True

        Returns
        -------
        out_str: str
            a string with all parameters from the info string
        """
        # init
        out = []
        # title
        if display_title:
            title = TITLE.format(self.__title)
            line = "─" * (len(title) - 1) + "\n"
            out.append(line + title + line)
        # params
        for section, elements in self.__elements.items():
            out.append(HEADER.format(section))
            for name, value in elements.items():
                if value is None:
                    out.append(PARAMSINGLE.format(name))
                else:
                    out.append(PARAM.format(name, value))
            # remove last an replace by a last param string
            out.pop()
            if value is None:
                out.append(LPARAMSINGLE.format(name))
            else:
                out.append(LPARAM.format(name, value))

        out_str = "".join(out)
        return out_str



"""Scipy-based integrators
============================

Implements integrator based on scipy's ``solve_ivp`` method.
"""

# % IMPORTS
import numpy as np
from scipy.integrate import solve_ivp
from functools import partial

# % LOCAL IMPORTS
from .simbase import Simulation, get_force_vec, _get_force_vec
from ..configurator import Configuration


# % SIMULATOR BASED ON SCIPY'S SOLVE_IVP


def get_force_vec_scipy(
    pos_speed_vector: np.ndarray, config: Configuration
) -> np.ndarray:
    """Computes the force on an atom, by adding all radiations pressures,
    in a Scipy compatible vectorization style

    Parameters
    ----------
    pos_speed_vector : array, shape (6,) or (6,k)
        cartesian position and speed vector
    config : Configuration
        a configuration object

    Returns
    -------
    force : array, shape (3,) or (3,k)
        the force at the coordinates given by ``pos_speed_vector``

    Notes
    -----

    The position/speed vector 'pos_speed_vector' should be of shape (6,) or (6,k)
    with the first dimension containing (x, y, z, vx, vy, vz) in the lab frame

    Note
    ----
        The function is vectorized to be compatible with Scipy's ``solve_ivp``
        function. Hence, it does not satisfy the functionnal vectorization
        used in the rest of this module

    Examples
    ---------

    .. code-block:: python

        # ... init a config object with the `Configuration` class
        from atomsmltr.simulation.simulator import get_force_vec_scipy
        import numpy as np

        # - init a position & speed vector grid
        # vx spans from -10 to 30
        # x, y, z, vy, vz set to 0
        vx_list = np.linspace(-10, 30, 301)
        pos_speed_vector = np.array([(0,0,0,vx,0,0,) for vx in vx_list]).T

        # - compute the force
        force = get_force_vec_scipy(pos_speed_vector, config)
        FX, FY, FZ = force

        # - print shapes for illustration
        print(f"{FX.shape=}")
        print(f"{pos_speed_vector.shape=}")
        print(f"{force.shape=}")


    This returns

    .. code-block:: python

        FX.shape=(301,)
        pos_speed_vector.shape=(6, 301)
        force.shape=(3, 301)


    """
    # TODO should we move that to the Configuration class ???
    # - get position and speed
    position = pos_speed_vector[0:3, ...].T
    speed = pos_speed_vector[3:6, ...].T
    # - compute force
    force = _get_force_vec(position, speed, config)
    # - transpose to satisfy vectorization rules
    force = force.T
    return force


def stop_position_event_scipy(
    t: float, u: np.ndarray, stop_position: list, offset: float = 0.0
):
    """Implements 'stop' events for Scipy's solve_ivp, based on atom's position

    Parameters
    ----------
    t : float
        time, not used here but required for the ``events`` functions in ``solve_ivp``
    u : array, shape (6,k)
        position/speed vector, according to ``solve_ivp`` vectorization convention
    stop_position : list
        list of Zones objects targetting position with actions set to stop

    Returns
    -------
    res: bool
        whether to stop the simulation

    See also
    --------
    atomsmltr.environment.zones
    atomsmltr.simulation.configurator.Configuration.get_stop_zones()
    """
    position = u[0:3, ...].T
    res = np.logical_and.reduce([zone.get_value(position) for zone in stop_position])
    res = res + offset
    return res


def stop_speed_event_scipy(
    t: float, u: np.ndarray, stop_speed: list, offset: float = 0.0
):
    """Implements 'stop' events for Scipy's solve_ivp, based on atom's speed

    Parameters
    ----------
    t : float
        time, not used here but required for the ``events`` functions in ``solve_ivp``
    u : array, shape (6,k)
        position/speed vector, according to ``solve_ivp`` vectorization convention
    stop_speed : list
        list of Zones objects targetting speed with actions set to stop

    Returns
    -------
    res: bool
        whether to stop the simulation

    See also
    --------
    atomsmltr.environment.zones
    atomsmltr.simulation.configurator.Configuration.get_stop_zones()
    """
    speed = u[3:6, ...].T
    res = np.logical_and.reduce([zone.get_value(speed) for zone in stop_speed])
    res = res + offset
    return res


class ScipyIVP_3D(Simulation):
    """A simulation class based on Scipy's ``solve_ivp`` solver

    Parameters
    ----------
    config : Configuration, optional
        the configuration to consider for the simulation
    method : str, optional
        method used for the ``solve_ivp`` solver, by default "Radau"
    **solve_ivp_args
        all other arguments are directly passed to ``solve_ivp``

    References
    ----------
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.solve_ivp.html

    """

    def __init__(
        self, config: Configuration = None, method: str = "Radau", **solve_ivp_args
    ):
        super(ScipyIVP_3D, self).__init__(config)
        self.solve_ivp_args = solve_ivp_args
        self.method = method

    # -- REQUESTED FUNCTIONS
    def _get_force_scipy(self, u):
        force = get_force_vec_scipy(u, self.config)
        return force

    def get_force(self, u):
        force = get_force_vec(u, self.config)
        return force

    def dudt(self, t, u):
        F = self._get_force_scipy(u)
        _, _, _, vx, vy, vz = u
        dx, dy, dz = vx, vy, vz
        dvx, dvy, dvz = F / self.config.atom.mass
        res = np.array([dx, dy, dz, dvx, dvy, dvz])
        return res

    def _integrate(self, u0, t):
        # - u0 to array
        u0 = np.asanyarray(u0)
        # - get stop events
        events = []
        stop_position, stop_speed = self.config.get_stop_zones()
        if stop_position:
            stop_pos = partial(
                stop_position_event_scipy, stop_position=stop_position, offset=-0.5
            )
            stop_pos.terminal = True
            events.append(stop_pos)
        if stop_speed:
            stop_sp = partial(
                stop_speed_event_scipy, stop_speed=stop_speed, offset=-0.5
            )
            stop_sp.terminal = True
            events.append(stop_sp)
        # - time
        t = np.asanyarray(t)
        if not t.shape:
            t = np.asanyarray([0, t])
        t = np.sort(t)
        t_span = (t[0], t[-1])
        # - integrate
        res = solve_ivp(
            fun=self.dudt,
            t_span=t_span,
            y0=u0,
            method=self.method,
            t_eval=t,
            events=events,
            **self.solve_ivp_args,
        )
        return res

    def _stop_event_speed(self, t, u):
        pass

    def _stop_event_position(self, t, u):
        pass

    def _u0_list_checker(self, value):
        if not hasattr(value, "__iter__"):
            raise ValueError("'u0_list' should be an iterable object")
        if value:
            for u0 in value:
                if np.asanyarray(u0).shape != (6,):
                    raise ValueError("'u0_list' should be a list of arrays of size 6")
        return value



"""
Examples : Configuration (1,1,1)
=======================

This example provides the configuration for the (1,1,1) 3D MOT as described
in http://dx.doi.org/10.1103/PhysRevLett.100.050801
"""

# % EXPERIENCE DESCRIPTION
description = """

Atomic fountain with launched 87Rb atoms in a (1,1,1) MOT configuration:

Ref: http://dx.doi.org/10.1103/PhysRevLett.100.050801

In a 3D magneto-optical trap (MOT), the standard setup uses three orthogonal pairs of
counter-propagating laser beams along the x, y, and z axis. If we picture the trapping zone as a cube, then
the (1,1,1) configuration is the same as a classical MOT, only rotated such that the summit initially
on (1,1,1) and the one on (0,0,0) are now both located on the z-axis.

To do so, we first rotate the lasers by φ = -π/4 (around the z-axis), thus the summit initially in (1,1,1) is
now in (sqrt(2), 0, 1).
Then, we rotate them by an angle θ around the y-axix, such that the the summit initially in (sqrt(2), 0, 1)
is now in (0, 0, sqrt(3)) (because the rotations are orthogonal, the norm of the vector remains the same),
and we solve for θ using the rotation matrix.

You can refer to the github in "" for exact calculations.

This simulation models an atomic fountain using cold 87Rb atoms initially trapped in a 3D
magneto-optical trap (MOT) in a configuration (1,1,1).
After cooling, the magnetic field is suppressed and the atomic ensemble is launched upwards along
the vertical axis using a moving optical molasses, which relies on a detuning (epsilon)
between the upward and the downward-propagating laser beams.

Physical configuration:
- Atom species : 87Rb (D2 cooling transition at 780 nm)
- Polarization : 2 counter-propagating σ- beams + 4 counter-propating σ+ beams
- Laser power : 16.7 mW
- Beam waist (1/e radius) : 15.5 mm
- MOT detuning : -3 Γ
- Detuning between upwards and donwards-propagating lasers : 1 Mhz

"""

# % IMPORTS

import numpy as np

# % LOCAL IMPORTS

from atomsmltr.environment import GaussianLaserBeam
from atomsmltr.simulation import Configuration
from atomsmltr.atoms import Rubidium
from atomsmltr.environment.lasers import CircularLeft, CircularRight
from atomsmltr.environment import ConstantForce

# --------------------------------------------------------------------------------------------------------


# % GENERATE CONFIGURATION of the (1,1,1) 3D MOT


# -- init config with rubidium atom
atom = Rubidium()

# -- get rubidium main (D2) transition information
main = atom.trans["main"]
gamma = main.Gamma


# -- setup magnetic field
# no magnetic field here ¯\_(ツ)_/¯

# -- add the relevant constant forces

m = Rubidium().mass  # kg
g = 9.81  # m/s^2
direction = np.array([0, 0, -1])  # along -z
grav_force = m * g * direction
gravity = ConstantForce(field_value=grav_force, tag="gravity")


# -- setup lasers of the 1D MOT
# cf. config from 'insert ref here'

laser_1 = GaussianLaserBeam(
    wavelength=780.241e-9,
    waist=22e-3,
    power=100e-3 / 6,
    waist_position=(0, 0, 0),
    direction_type="thetaphi",
    direction=(0.95500000000000, 3.141592653589793),
    polarization=CircularLeft(),
    tag="las1",
)

laser_2 = GaussianLaserBeam(
    wavelength=780.241e-9,
    waist=22e-3,
    power=100e-3 / 6,
    waist_position=(0, 0, 0),
    direction_type="thetaphi",
    direction=(2.186592653589793, 0),
    polarization=CircularLeft(),
    tag="las2",
)

laser_3 = GaussianLaserBeam(
    wavelength=780.241e-9,
    waist=22e-3,
    power=100e-3 / 6,
    waist_position=(0, 0, 0),
    direction_type="thetaphi",
    direction=(0.9554749537638146, -1.047003706391126),
    polarization=CircularRight(),
    tag="las3",
)

laser_4 = GaussianLaserBeam(
    wavelength=780.241e-9,
    waist=22e-3,
    power=100e-3 / 6,
    waist_position=(0, 0, 0),
    direction_type="thetaphi",
    direction=(2.186117699825979, 2.0945889471986674),
    polarization=CircularRight(),
    tag="las4",
)

laser_5 = GaussianLaserBeam(
    wavelength=780.241e-9,
    waist=22e-3,
    power=100e-3 / 6,
    waist_position=(0, 0, 0),
    direction_type="thetaphi",
    direction=(0.9554749537638146, 1.047003706391126),
    polarization=CircularRight(),
    tag="las5",
)

laser_6 = GaussianLaserBeam(
    wavelength=780.241e-9,
    waist=22e-3,
    power=100e-3 / 6,
    waist_position=(0, 0, 0),
    direction_type="thetaphi",
    direction=(2.186117699825979, -2.0945889471986674),
    polarization=CircularRight(),
    tag="las6",
)


# -- add everything to the configuration
config = Configuration()
config.atom = atom
config.add_objects([gravity])
config += laser_1, laser_2, laser_3, laser_4, laser_5, laser_6


# -- setup lasers detuning parameters of the 3D MOT
epsilon = 2 * np.pi * 1e6
detuning = -3 * gamma


# -- add the detuning and the right epsilon to the atomlight coupling

list_lasers = config.list_lasers()
for laser_name in list_lasers:
    laser = config.get_laser_copy(laser_name)
    direction = laser.direction
    if direction[0] < np.pi / 2:
        config.add_atomlight_coupling(laser_name, "main", detuning + epsilon)
    else:
        config.add_atomlight_coupling(laser_name, "main", detuning - epsilon)



"""configuration
==================

Here we implement the ``Configuration`` class, that allows to define
consistent configurations that are later fed to the ``Simulation`` objects.

Quick description
--------------------

A configuration consists of:

* a atom (``Atom``)
* a collection of laser beams (``LaserBeam``)
* a collection of magnetic Fields (``MagneticField``)
* a collection of forces (``Force``)
* a collection of zones (``Zone``)

The coupling between atoms and lasers is stored in a ``atomlight`` dictionnary, that
is setup with the ``add_atomlight_coupling()`` method.
"""

# % IMPORTS
import warnings
import numpy as np
from copy import copy, deepcopy

# % LOCAL IMPORTS
from ..environment import LaserBeam, MagneticField, Zone, SuperZone, Force
from ..environment.envbase import EnvObject
from ..atoms import Atom
from ..utils.infostring import InfoString

# % CONSTANTS

SEP_STR = "# ------------ {} ------------ #"

# % DEFINE THE CLASS


class Configuration(object):
    """Defines a configuration for the simulation

    Parameters
    ----------
    object_list : EnvObject | list, optional
        list of environment objects (lasers, magnetic fields, zones, forces)
        to include in the configuration, by default None
    atom : Atom, optional
        atom for the simulation, by default None

    Examples
    ---------

    .. code-block:: python

        # - imports
        from atomsmltr.environment import GaussianLaserBeam, MagneticOffset, Limits
        from atomsmltr.atoms import Ytterbium
        from atomsmltr.simulation import Configuration

        # - setup environment objects
        laser = GaussianLaserBeam(tag="laser")
        mag_offset = MagneticOffset(offset=(0,1,0), tag="offset")
        xlim = Limits(min=0, max=10, axis=0, target="position")

        # - setup atom
        yb = Ytterbium()

        # - init configuration
        config = Configuration(object_list=[laser, mag_offset, xlim], atom=yb)

        # - print info
        config.print_info()


    """

    def __init__(self, object_list: EnvObject | list = None, atom: Atom = None):

        # - initialize collections
        self.__lasers = {}
        self.__zones = {}
        self.__magfields = {}
        self.__forces = {}
        self.__atomlight = {}
        self.__atom = None

        self.__implemented_collections = {
            "laser": self.__lasers,
            "magnetic field": self.__magfields,
            "force": self.__forces,
            "zone": self.__zones,
        }

        # - init atom
        if atom is not None:
            self.atom = atom

        if object_list is not None:
            self.add_objects(object_list)

    # -- ATOM-LIGHT INTERACTION HANDLING

    def get_atomlight_couples(self) -> list:
        """Returns a list of (transition, laser, detuning) tuples

        Returns
        -------
        list
            a list of tuples with (transition, laser, detuning)
        """
        list = []
        for transition_tag, laser_dict in self.__atomlight.items():
            transition = self.atom.trans[transition_tag]
            for laser_tag, coupling_info in laser_dict.items():
                laser = self.__lasers[laser_tag]
                detuning = coupling_info["detuning"]
                list.append((transition, laser, detuning))
        return list

    def add_atomlight_coupling(
        self,
        laser: str | LaserBeam,
        transition: str,
        detuning: float,
        verbose: bool = False,
        override: bool = False,
    ):
        """Adds a atom-light coupling element in the configuration

        Parameters
        ----------
        laser : str | LaserBeam
            either a laser tag or a laser object. This object/tag has to be in the configuration laser list
        transition : str
            the tag of the transition. Should be part of the collection's atom transition list
        detuning : float
            detuning of the laser w.r.t to transition (rad/s), see notes for the definition
        verbose : bool, optional
            if True, will print messages when adding the couplint, by default False
        override : bool, optional
            if set to True, if a coupling between the laser and the transition already
            exists, then it will be overriden. Otherwise it will raise an error, by default False

        Notes
        ------

        The detuning δ is defined as:

            δ = ωL - ω0

        Where ωL is the laser pulsation and ω0 the atomic transition pulsation (hence, in rad/s)
        Stated otherwise, detuning units is in units of 2π x Hz
        """
        # - checking inputs
        # check laser argument
        if not isinstance(laser, (str, LaserBeam)):
            raise TypeError("'laser' should be a tag (string) or a Laser object")
        if not isinstance(laser, str):
            laser = laser.tag
        # check that laser is there
        if laser not in self.__lasers:
            msg = f"No entry for laser tag '{laser}'. "
            msg += f" Available lasers are {list(self.__lasers)}."
            raise KeyError(msg)
        # check that transition is there
        if self.atom is None:
            raise ValueError("No atom was defined for this config")
        if transition not in self.__atomlight:
            msg = f"No entry for transition '{transition}'. "
            msg += f" Available transitions are {list(self.__atomlight)}."
            raise KeyError(msg)

        # - check that there is no link
        if laser in self.__atomlight[transition]:
            msg = f"There is alreay a link between laser '{laser}' and transition '{transition}'. "
            if not override:
                msg += "Since 'override' is set to 'False', we stop here with an error."
                raise KeyError(msg)
            else:
                msg += "Since 'override' is set to 'True', we go on."
                if verbose:
                    print(" > " + msg)
        # - store
        self.__atomlight[transition][laser] = {"detuning": detuning}

    def rm_atomlight_coupling(
        self,
        laser: str | LaserBeam,
        transition: str,
    ):
        """Removes an atom-light coupling

        Parameters
        ----------
        laser : str | LaserBeam
            laser coupled : tag (str) or directly the object
        transition : str
            transition tag
        """
        # - checking inputs
        # check laser argument
        if not isinstance(laser, (str, LaserBeam)):
            raise TypeError("'laser' should be a tag (string) or a Laser object")
        if not isinstance(laser, str):
            laser = laser.tag

        # - remove
        success = False
        if transition in self.__atomlight:
            if laser in self.__atomlight[transition]:
                self.__atomlight[transition].pop(laser)
                success = True
        if not success:
            msg = f"There is no link between '{laser}' and '{transition}'."
            raise KeyError(msg)

    def reset_atomlight_coupling(self):
        for transition in self.__atomlight:
            self.__atomlight[transition].clear()

    # -- GETTING VALUES
    def getB(self, position: np.ndarray) -> np.ndarray:
        """Returns the total magnetic field at a given position in the lab frame

        Parameters
        ----------
        position : array, shape (3,) or (n1, n2, .., 3)
            array of cartesian coordinates in the lab frame

        Returns
        -------
        B : array, shape (3,) or (n1, n2, .., 3)
            magnetic field at the position. shape matches the one of ``position``

        Notes
        ------
        position is an array_like object, with shape (3,) or (n1, n2, .., 3).

        In all cases, the last dimension contains cordinates (x, y, z),
        in meter and in the lab frame

        Example
        -------

        .. code-block:: python

            ... init a proper config first
            import numpy as np

            # generate a grid of 100 x 100 points in the (x, y) plane
            grid = np.mgrid[-10:10:100j, -10:-10:100j, 0:0:1j]
            grid = np.squeeze(grid)

            # get coordinates arrays (for plotting for instance)
            X, Y, Z = grid

            # generate the requested (..., 3) shaped position array
            position = grid.T

            # compute magnetic field
            B = config.getB(position)

            # get magnetic field components
            Bx, By, Bz = B.T

            # show shapes (to illustrate what we did)
            print(f"{grid.shape=}")
            print(f"{position.shape=}")
            print(f"{B.shape=}")
            print(f"{X.shape=}")
            print(f"{Bx.shape=}")

        This returns

        .. code-block:: python

            grid.shape=(3, 100, 100)
            position.shape=(100, 100, 3)
            B.shape=(100, 100, 3)
            X.shape=(100, 100)
            Bx.shape=(100, 100)

        """
        B = np.zeros_like(position, dtype=float)
        if self.__magfields:
            for magfield in self.__magfields.values():
                B += magfield.get_value(position)
        return B

    def getBnorm(self, position):
        """Returns the magnetic field amplitude (norm) at a given lab position

        Parameters
        ----------
        position : array, shape (3,) or (n1, n2, .., 3)
            array of cartesian coordinates in the lab frame

        Returns
        -------
        B_norm : array, shape (1,) or (n1, n2, .., 1)
            magnetic field norm the position. shape matches the one of ``position``

        Notes
        ------
        position is an array_like object, with shape (3,) or (n1, n2, .., 3).

        In all cases, the last dimension contains cordinates (x, y, z),
        in meter and in the lab frame
        """
        B = self.getB(position)
        Bx, By, Bz = B.T
        B_norm = np.sqrt(Bx**2 + By**2 + Bz**2).T
        return B_norm

    # -- GETTING ZONES
    def get_stop_zones(self):
        """Returns two list of the zones whose ``action`` are set to ``stop``

        Returns
        -------
        stop_position: list
            list of position stop zones (target=position)
        stop_speed: list
            list of speed stop zones (target=speed)
        """

        return self._get_zones(action="stop")

    def get_all_zones(self):
        """Returns zones sorted in according to their target

        Returns
        -------
        stop_position: list
            list of position stop zones (target=position)
        stop_speed: list
            list of speed stop zones (target=speed)
        """

        return self._get_zones(action="all")

    def _get_zones(self, action: str = "stop"):
        """Returns two list of the zones whose ``action`` are set to a given value

        Parameters
        ----------
        action : str, optional
            action to target, by default "stop"

        Returns
        -------
        stop_position: list
            list of position zones (target=position)
        stop_speed: list
            list of speed zones (target=speed)
        """
        stop_speed = []
        stop_position = []
        for zone in self.__zones.values():
            if zone.action == action or action == "all":
                if zone.target == "speed":
                    stop_speed.append(deepcopy(zone))
                elif zone.target == "position":
                    stop_position.append(deepcopy(zone))
        return stop_position, stop_speed

    def in_zone(self, pos_speed_vector: np.ndarray, action: str = "stop") -> np.ndarray:
        """Evaluates whether 'pos_speed_vector' is in the zones corresponding to a given action

        Parameters
        ----------
        vector : array of shape (6,) or (n1, n2, ..., 6)
            cartesian coordinates of the vectors in the lab frame
        action : str, optionnal
            the action of the zones to consider by default "stop"

        Returns
        -------
        in_zone : array of shape (1,) or (n1, n2, ..., 1)
            whether the vector is 'in the zone'

        Notes
        -----

        ``vector`` should be an array of shape (6,) or (n1, n2, .., 6), where last axis contains
        the coordinates (position & speed) to evaluate.

        In all cases, the last dimension contains cordinates (x, y, z, vx, vy, vz),
        """

        # -- init a SuperZone
        collection = SuperZone(logic="AND")

        # -- populate
        for zone in self.__zones.values():
            if zone.action == action:
                collection += zone

        # -- return results
        return collection.get_value(pos_speed_vector)

    # -- GETTING FORCES

    def get_all_forces(self) -> list:
        """Returns a list of all forces

        Returns
        -------
        force_list (list)
            a list of all forces in the configuration
        """
        force_list = self.__forces.values()
        return force_list

    # -- COLLECTION HANDLING METHODS

    # ADDING
    def add_objects(self, obj: EnvObject | list, verbose=False):
        """Add environment objects to the configuration.

        Parameters
        ----------
        obj : EnvObject | list
            a environment object or a list of objects
        verbose : bool, optional
            if set to True messages are displayed. Defaults to False.

        Notes
        -----

        The function takes a single environment object (laser, magnetic field...) or a collection
        of objects in the form of a tuple or a list.

        Objects of different subtypes can be added at the same time: the method will add them
        to the correct collection based on their classes

        Note
        ----
            The addition operator ``+`` also allows to add objects.
            Hence, ``conf.add_objects([obj1, obj2, ...])`` is equivalent to
            ``conf += obj1, obj2 , ...``

        Examples
        --------
        .. code-block:: python

            ... init a proper config first and env objects

            # add objects
            config.add_objects(laser1)
            config.add_objects([mag_field, zone1, zone2, laser2])

            # also works with += operator*
            config += laser3, laser4

        """

        # - check argument
        self.__check_objects_arg(obj)

        # - recursive add if list
        if isinstance(obj, (list, tuple)):
            for element in obj:
                self.add_objects(element, verbose)
            return

        # - add object
        if isinstance(obj, MagneticField):
            collection = self.__magfields
            name = "magnetic fields"
        elif isinstance(obj, LaserBeam):
            collection = self.__lasers
            name = "lasers"
        elif isinstance(obj, Force):
            collection = self.__forces
            name = "forces"
        elif isinstance(obj, Zone):
            collection = self.__zones
            name = "zones"
        else:
            msg = f"Objects of type {type(obj)} are not handled yet.. where did you find this ?"
            raise TypeError(msg)
        self.__add_obj(obj, collection, name)

        if verbose:
            msg = f"(+) sucessfully added object '{obj.tag}' in the {name} collection"
            print(msg)

    def __add_obj(self, obj, collection, name):
        """Internal method to add objects"""
        # - copy
        obj = copy(obj)
        # - check that object tag not present
        msg = f"We already have an element with tag '{obj.tag}' in our {name} collection. "
        msg += "Remove or update this element."
        if obj.tag in collection:
            raise ValueError(msg)
        # - add the object >>> we use a copy to avoid unwanted modifications
        collection[obj.tag] = obj

    # UPDATING
    def update_objects(self, obj: EnvObject | list, verbose=False, error_on_fail=False):
        """Update an object or a list of objects

        Parameters
        ----------
        obj : EnvObject | list
            a environment object or a list of objects
        verbose : bool, optional
            if set to True messages are displayed. Defaults to False.
        error_on_fail : bool, optional
            if set to True, raises an error if it fails. Otherwise,
            just raises a warning and continues. Defaults to False.

        Notes
        ------
        The function takes a single environment object (laser, magnetic field...) or a collection
        of objects in the form of a tuple or a list.

        For each object given as an input, if there is an object with:

        (1) same type (laser, magnetic field) **and**
        (2) same tag

        then this object is replaced by the new one.
        """
        # - check argument
        self.__check_objects_arg(obj)

        # - recursive add if list
        if isinstance(obj, (list, tuple)):
            for element in obj:
                self.update_objects(element, verbose, error_on_fail)
            return
        # - add object
        if isinstance(obj, MagneticField):
            collection = self.__magfields
            name = "magnetic fields"
        elif isinstance(obj, Force):
            collection = self.__forces
            name = "forces"
        elif isinstance(obj, Zone):
            collection = self.__zones
            name = "zones"
        elif isinstance(obj, LaserBeam):
            collection = self.__lasers
            name = "lasers"
        else:
            msg = f"Objects of type {type(obj)} are not handled yet.. where did you find this ?"
            raise TypeError(msg)
        success = self.__upd_obj(obj, collection, name, error_on_fail)
        if verbose:
            if success:
                msg = f"(>) sucessfully updated object '{obj.tag}' in the {name} collection"
            else:
                msg = f"(x) could not update '{obj.tag}' in the {name} collection"
            print(msg)

    def __upd_obj(self, obj, collection, name, error_on_fail) -> bool:
        # - copy
        obj = copy(obj)
        # - check that object tag not present
        msg = f"There is no element with tag '{obj.tag}' in our {name} collection. "
        if not obj.tag in collection:
            if error_on_fail:
                raise KeyError(msg)
            else:
                warnings.warn(msg)
            return False
        # - update the object
        collection[obj.tag] = obj
        return True

    # LISTING
    def list_lasers(self) -> list:
        """Returns the list of laser's tags in the current config"""
        return list(self.__lasers)

    def list_magnetic_fields(self):
        """Returns the list of magnetic fields' tags in the current config"""
        return list(self.__magfields)

    def list_zones(self):
        """Returns the list of zones' tags in the current config"""
        return list(self.__zones)

    def list_forces(self):
        """Returns the list of forces' tags in the current config"""
        return list(self.__forces)

    # REMOVING
    def rm_object(self, collection: str, tag: str):
        """Remove object from 'collection' with 'tag'

        Collection must be in ['laser', 'magnetic field', 'zone', 'force']

        Parameters
        ----------
        collection : str
            the collection from which the object should be removed
        tag : str
            the tag of the object
        """

        coll = self.__check_object_in_coll(collection, tag)
        del coll[tag]

    def rm_laser(self, tag: str):
        """Removes laser by tag

        Parameters
        ----------
        tag : str
            laser tag
        """
        return self.rm_object("laser", tag)

    def rm_magnetic_field(self, tag):
        """Removes magnetic field by tag

        Parameters
        ----------
        tag : str
            magnetic field tag
        """
        return self.rm_object("magnetic field", tag)

    def rm_zone(self, tag):
        """Removes zone by tag

        Parameters
        ----------
        tag : str
            zone tag
        """
        return self.rm_object("zone", tag)

    def rm_force(self, tag):
        """Removes force by tag

        Parameters
        ----------
        tag : str
            force tag
        """
        return self.rm_object("force", tag)

    def rm_all_objects(self):
        """Remove all objects"""
        self.rm_all_lasers()
        self.rm_all_magnetic_fields()
        self.rm_all_zones()
        self.rm_all_forces()

    def rm_all_lasers(self):
        """Remove all lasers"""
        self.__lasers.clear()

    def rm_all_magnetic_fields(self):
        """Remove all magnetic fields"""
        self.__magfields.clear()

    def rm_all_zones(self):
        """Remove all zones"""
        self.__zones.clear()

    def rm_all_forces(self):
        """Remove all forces"""
        self.__forces.clear()

    # -- INFOS

    def gen_object_infostring_object(self, collection: str, tag: str) -> InfoString:
        """Generate infostring object for an object from 'collection' with 'tag'

        Collection must be in ['laser', 'magnetic field', 'zone']

        Parameters
        ----------
        collection : str
            the collection from which the object should be taken
        tag : str
            the tag of the object

        Returns
        -------
        infostring: Infostring
            an infostring object

        See also
        ---------
        atomsmltr.utils.infostring
        """

        coll = self.__check_object_in_coll(collection, tag)
        info = coll[tag].gen_infostring_obj()
        info.title = f"{collection} | {tag=}"
        return info

    def print_object_info(self, collection: str, tag: str):
        """Print info for an object from 'collection' with 'tag'

        Collection must be in ['laser', 'magnetic field', 'zone'    ]

        Parameters
        ----------
        collection : str
            the collection from which the object should be taken
        tag : str
            the tag of the object
        """
        info = self.gen_object_infostring_object(collection, tag)
        print(info.generate())

    def print_laser_info(self, tag: str):
        """Print info of the laser indentified by 'tag'

        Parameters
        ----------
        tag : str
            the tag of the laser
        """
        return self.print_object_info("laser", tag)

    def print_magnetic_field_info(self, tag: str):
        """Print info of the magnetic field indentified by 'tag'

        Parameters
        ----------
        tag : str
            the tag of the magnetic field
        """
        return self.print_object_info("magnetic field", tag)

    def print_zone_info(self, tag: str):
        """Print info of the zone indentified by 'tag'

        Parameters
        ----------
        tag : str
            the tag of the zone
        """
        return self.print_object_info("zone", tag)

    def print_force_info(self, tag: str):
        """Print info of the force indentified by 'tag'

        Parameters
        ----------
        tag : str
            the tag of the force
        """
        return self.print_object_info("force", tag)

    # -- GET OBJECTS
    def get_object_copy(self, collection: str, tag: str) -> EnvObject:
        """Returns a copy of an object from 'collection' with 'tag'

        Collection must be in ['laser', 'magnetic field', 'zone', 'force' ]

        Parameters
        ----------
        collection : str
            the collection from which the object should be taken
        tag : str
            the tag of the object
        """
        coll = self.__check_object_in_coll(collection, tag)
        return copy(coll[tag])

    def get_laser_copy(self, tag: str):
        """Returns a copy of the laser indentified by 'tag'

        Parameters
        ----------
        tag : str
            the tag of the laser
        """
        return self.get_object_copy("laser", tag)

    def get_magnetic_field_copy(self, tag: str):
        """Returns a copy of the magnetic field indentified by 'tag'

        Parameters
        ----------
        tag : str
            the tag of the magnetic field
        """
        return self.get_object_copy("magnetic field", tag)

    def get_zone_copy(self, tag: str):
        """Returns a copy of the zone indentified by 'tag'

        Parameters
        ----------
        tag : str
            the tag of the zone
        """
        return self.get_object_copy("zone", tag)

    def get_force_copy(self, tag: str):
        """Returns a copy of the force indentified by 'tag'

        Parameters
        ----------
        tag : str
            the tag of the force
        """
        return self.get_object_copy("force", tag)

    # -- COMMON METHODS

    def __check_object_in_coll(self, collection, tag) -> dict:
        implemented_collections = self.__implemented_collections
        if collection not in implemented_collections:
            msg = f"Wrong collection. implemented collections are {list(implemented_collections)}"
            raise ValueError(msg)

        coll = implemented_collections[collection]
        if tag not in coll:
            msg = f"There is no {collection} with tag {tag}"
            raise KeyError(msg)
        return coll

    def __check_objects_arg(self, obj):
        """Called in add_objects & update_objects"""
        type_err_msg = "passed argument should be an EnvObject or a list of EnvObjects"
        if isinstance(obj, (list, tuple)):
            for element in obj:
                if not isinstance(element, EnvObject):
                    raise TypeError(type_err_msg)
        elif not isinstance(obj, EnvObject):
            raise TypeError(type_err_msg)

    # -- GETTERS & SETTERS
    @property
    def atom(self) -> Atom:
        """ "Atom: the configuration atom"""
        return self.__atom

    @atom.setter
    def atom(self, atom: Atom):
        # - set atom
        if not isinstance(atom, Atom):
            raise TypeError("'atom' should be an atom")
        self.__atom = atom
        # - prepare atomlight dict
        # issue warning if already some entries
        if not self.__atomlight:
            # if dict not empty, clear it
            self.__atomlight.clear()
            # warnings.warn("Resetting atom-light dictionnary...")
        for transition_tag in self.atom.list_transitions():
            self.__atomlight[transition_tag] = {}

    @property
    def objects(self) -> dict:
        """dict: the collection of objects"""
        out = {}
        for k, v in self.__implemented_collections.items():
            out[k] = copy(v)
        return out

    # -- INFO PRINTER
    def gen_atomlight_infostring_obj(self):
        info = InfoString("Atom-light couplings")
        for transition, couplings in self.__atomlight.items():
            info.add_section(f"transition > '{transition}'")
            if couplings:
                for laser, params in couplings.items():
                    detuning = params["detuning"]
                    trans_Gamma = self.atom.trans[transition].Gamma
                    det_str = f"{detuning=:.3g}"
                    det_str += f" ({detuning / trans_Gamma:.2f}Γ)"
                    info.add_element(f"laser '{laser}'", det_str)
            else:
                info.add_element("empty")
        return info

    def print_atomlight_info(self):
        """Prints atom-light coupling information"""
        print(self.gen_atomlight_infostring_obj().generate())

    def gen_infostring_obj_list(self):
        # - prepare output
        info_list = []
        # - general infostring
        info = InfoString("General informations")
        # atom
        info.add_section("atom")
        info.add_element("name", self.atom.name)
        # collections
        for name, coll in self.__implemented_collections.items():
            info.add_section(name + "s")
            if coll:
                for tag in coll:
                    info.add_element(tag)
            else:
                info.add_element("empty")

        # append to list
        info_list.append(info)

        # - atom info
        info = self.atom.gen_infostring_obj()
        info.title = f"atom | {self.atom.name.lower()}"
        info_list.append(info)

        # - collections
        for name, coll in self.__implemented_collections.items():
            for tag in coll:
                info = self.gen_object_infostring_object(name, tag)
                info_list.append(info)

        # - atom light
        info = self.gen_atomlight_infostring_obj()
        info_list.append(info)

        return info_list

    def print_info(self):
        """Prints informations on the configuration"""
        info_list = self.gen_infostring_obj_list()
        print(SEP_STR.format("CONFIG INFO > START"))
        for info in info_list:
            print(info.generate())
        print(SEP_STR.format("CONFIG INFO > STOP "))

    # -- OPERATORS OVERLOADING

    def __add__(self, object: EnvObject):
        """returns a copy of the configuration, with the object added"""
        new_config = deepcopy(self)
        new_config.add_objects(object)
        return new_config

    def __iadd__(self, object: EnvObject):
        """adds the objects"""
        self.add_objects(object)
        return self

    def __imod__(self, object: EnvObject):
        """updates the objects"""
        self.update_objects(object)
        return self



"""
magnetic fields
=======================

This module implements Magnetic field classes, which are mostly bare implementations of
generic fields defined in ``atomsmltr.environment.fields``.

See also
---------
atomsmltr.environment.fields.generic
atomsmltr.environment.fields.interpolated
"""

# % LOCAL IMPORTS
from ..generic import (
    Field,
    GradientField,
    ConstantField,
    QuadrupoleFieldX,
    QuadrupoleFieldZ,
    QuadrupoleFieldY,
    QuadrupoleField,
)
from ..interpolated import InterpolatedField1D1D, InterpolatedField3D3D

# % CLASSES

# -- MAG FIELDs PARENT CLASS
#   not really used currently, but will be useful if we need to
#   implement features specific to mag. fields.
# > will also allow to check that the field is indeed a magnetic field
#   in the configuration object


class MagneticField(Field):
    """A generic magnetic field class. Used to set some properties common to all magnetic
    fields objects, and to have a way to identify magnetic field objects."""

    @property
    def type(self):
        return "magnetic field"

    @property
    def unit(self):
        return "T"


# -- PERFECT FIELDS CLASSES
class MagneticOffset(MagneticField, ConstantField):
    """A perfect magnetic field offset

    See also
    ---------
    atomsmltr.environment.fields.generic.ConstantField
    """

    @property
    def type(self):
        return "magnetic field offset"


class MagneticGradient(MagneticField, GradientField):
    """A perfect magnetic field gradient

    See also
    ---------
    atomsmltr.environment.fields.generic.GradientField
    """

    @property
    def type(self):
        return "magnetic field gradient"


class MagneticQuadrupoleX(MagneticField, QuadrupoleFieldX):
    """A perfect magnetic field quadrupole, with strong axis along X

    See also
    ---------
    atomsmltr.environment.fields.generic.QuadrupoleFieldX
    """

    @property
    def type(self):
        return "magnetic quadrupole x"


class MagneticQuadrupoleY(MagneticField, QuadrupoleFieldY):
    """A perfect magnetic field quadrupole, with strong axis along Y

    See also
    ---------
    atomsmltr.environment.fields.generic.QuadrupoleFieldY
    """

    @property
    def type(self):
        return "magnetic quadrupole y"


class MagneticQuadrupoleZ(MagneticField, QuadrupoleFieldZ):
    """A perfect magnetic field quadrupole, with strong axis along Z

    See also
    ---------
    atomsmltr.environment.fields.generic.QuadrupoleFieldZ
    """

    @property
    def type(self):
        return "magnetic quadrupole z"


class MagneticQuadrupole(MagneticField, QuadrupoleField):
    """A perfect magnetic field quadrupole, with strong axis along a given vector (x,y,z)

    See also
    ---------
    atomsmltr.environment.fields.generic.QuadrupoleField
    """

    @property
    def type(self):
        return "magnetic quadrupole"


# -- INTERPOLATED
class InterpMag1D1D(MagneticField, InterpolatedField1D1D):
    """An interpolated field 1D / 1D

    See also
    ---------
    atomsmltr.environment.fields.interpolated.InterpolatedField1D1D
    """

    @property
    def type(self):
        return "interpolated mag. field (1D-1D)"


# -- INTERPOLATED
class InterpMag3D3D(MagneticField, InterpolatedField3D3D):
    """An interpolated field 3D/3D

    See also
    ---------
    atomsmltr.environment.fields.interpolated.InterpolatedField3D3D
    """

    @property
    def type(self):
        return "interpolated mag. field (3D-3D)"



"""Ytterbium
=============
Implements a dedicated class for Ytterbium atoms

>>> from atomsmltr.atoms import Ytterbium
"""

# % IMPORTS
import scipy.constants as csts

# % LOCAL IMPORTS
from ...atoms.generic import Atom
from ...atoms.transitions import J0J1Transition

# % CONSTANTS

YTTERBIUM_174_MASS = 173.94 * csts.m_u  # kg
"""float: Ytterbium 88 mass (kg)"""

YTTERBIUM_MAIN_WAVELENGTH = 398.911e-9  # m
"""float: Ytterbium main transition wavelength (m)"""
YTTERBIUM_MAIN_GAMMA = 2 * csts.pi * 28.9e6  # rad/s
"""float: Ytterbium main transition natural linewidth (rad/s)"""
YTTERBIUM_MAIN_LANDE_FACTOR = 1.035
"""float: Ytterbium main transition Lande factor"""

YTTERBIUM_INTERCOMBINATION_WAVELENGTH = 555.802e-9  # m
"""float: Ytterbium intercombination transition wavelength (m)"""
YTTERBIUM_INTERCOMBINATION_GAMMA = 2 * csts.pi * 182e3  # rad/s
"""float: Ytterbium intercombination transition natural linewidth (rad/s)"""
YTTERBIUM_INTERCOMBINATION_LANDE_FACTOR = 1.493
"""float: Ytterbium intercombination transition Lande factor"""

# %% TRANSITIONS


class MainLine(J0J1Transition):
    """The main (399nm) transition of Ytterbium"""

    def __init__(self):
        super().__init__(
            lande_factor=YTTERBIUM_MAIN_LANDE_FACTOR,
            wavelength=YTTERBIUM_MAIN_WAVELENGTH,
            Gamma=YTTERBIUM_MAIN_GAMMA,
            tag="main",
        )


class IntercombinationLine(J0J1Transition):
    """The intercombination (556nm) transition of Ytterbium"""

    def __init__(self):
        super().__init__(
            lande_factor=YTTERBIUM_INTERCOMBINATION_LANDE_FACTOR,
            wavelength=YTTERBIUM_INTERCOMBINATION_WAVELENGTH,
            Gamma=YTTERBIUM_INTERCOMBINATION_GAMMA,
            tag="intercombination",
        )


# %% ATOM


class Ytterbium(Atom):
    """Strontium 174 atomic class"""

    def __init__(self):

        # init super class
        super().__init__(
            mass=YTTERBIUM_174_MASS,
            name="Ytterbium",
        )
        # add transitions
        main = MainLine()
        intercomb = IntercombinationLine()
        self.add_transition(main)
        self.add_transition(intercomb)



"""The ``atomsmltr.environment.zones`` subpackage provides definitions for spatial zones

Those zones are then used in ``Configuration`` objects to define position and speed zones
in which some actions have to be taken - for instance, stopping the simulation

Create a zone
---------------

Define two zones, for x and vx

.. code-block:: python

    from atomsmltr.environment.zones import Limits

    xlim = Limits(min=-1, max=1, axis=0, target="position", action="stop", tag="xlim")
    vxlim = Limits(min=0, max=500, axis=0, target="speed", action="stop", tag="vxlim")


Combining Zones
---------------

Once a series of ``Zone`` are defined, it is possible to combine them using
basic python operators. The result will be a ``ZoneCollection`` with different
combination rules, depending on the operator.

Note that a ``ZoneCollection`` is a particular kind of ``Zone``, so operators
also work on ``ZoneCollection``.


ANDCollection
~~~~~~~~~~~~~~~~

A ``ANDCollection`` can be generated using the ``&`` operator :

>>> and_collection = zone1 & zone2

The created ``and_collection`` object then has two zones in its ``.zones`` list,
namely ``zone1`` and ``zone2``. Its ``in_zone()`` method will return the result of
``zone1.get_value() and zone2.get_value()``


ORCollection
~~~~~~~~~~~~~~~~

A ``ORCollection`` can be generated using the ``|`` operator :

>>> or_collection = zone1 | zone2

XORCollection
~~~~~~~~~~~~~~~~

A ``XORCollection`` can be generated using the ``^`` operator :

>>> xor_collection = zone1 ^ zone2

Adding zones to a collection
~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Zones can be added to a collection using the ``+=`` operator :

>>> xor_collection = zone1 ^ zone2
>>> xor_collection += zone3

Zones of a same kind can also be appended

>>> and_coll1 = zone1 & zone2
>>> and_coll2 = zone3 & zone4
>>> and_coll3 = and_coll1 + and_coll2


See also
--------
atomsmltr.environment.zones.generic
atomsmltr.environment.zones.limits
atomsmltr.environment.zones.volumes

"""

__all__ = [
    "UpperLimit",
    "LowerLimit",
    "Limits",
    "Box",
    "Cylinder",
    "Zone",
    "SuperZone",
]

from .limits import UpperLimit, LowerLimit, Limits
from .volumes import Box, Cylinder
from .generic import Zone, SuperZone



"""The ``atomsmltr.utils.plotter`` subpackage defines the ``Plottable`` class.

See also
---------
atomsmltr.utils.plotter.plotter

"""

__all__ = [
    "Axes3D",
    "Plottable",
]

from .tools import Axes3D
from .plotter import Plottable



"""The ``atomsmltr.simulation`` subpackage provides classes to manage
configurations and simulations

See also
---------
atomsmltr.simulation.configurator
atomsmltr.simulation.simulator
"""

__all__ = [
    "Configuration",
    "ScipyIVP_3D",
    "RK4",
    "RK4St",
    "Euler",
    "EulerSt",
    "VelocityVerlet",
]

from .configurator import Configuration
from .simulator import (
    ScipyIVP_3D,
    RK4,
    RK4St,
    Euler,
    EulerSt,
    VelocityVerlet,
)



"""Home-made stochastic integrators
=========================================

Implements homemade integrators for stochastic systems, that is, taking into account
fluctuations due to photon scattering
"""

# % IMPORTS
import numpy as np
import scipy.constants as csts
from functools import partial

# % LOCAL IMPORTS
from .simbase import Simulation, SimRes, get_force_vec
from .deterministic import CustomSimulationBase
from ..configurator import Configuration


# % USEFUL FUNCTIONS
def random_unit_vector(shape=(1,)):
    """Generates a random unit vector

    Parameters
    ----------
    shape : tuple, optional
        shape of the output will be (**shape, 3), by default (1,)

    Returns
    -------
    vec : array
        the random unit vector
    """
    # - get random phi and costheta
    rng = np.random.default_rng()
    phi = rng.uniform(low=0, high=2 * np.pi, size=shape)
    costheta = rng.uniform(low=-1, high=1, size=shape)
    sintheta = np.sqrt(1 - costheta**2)
    # - compute x, y, z
    x = sintheta * np.cos(phi)
    y = sintheta * np.sin(phi)
    z = costheta
    # - combine into an array of good shape
    vec = np.array([x.T, y.T, z.T]).T
    return vec


# % HOME-MADE SIMULATORS


class RK4St(CustomSimulationBase):
    """A homemade simulator based on fourth order Runge-Kutta method, taking into
    account spontaneous emission.

    Parameters
    ----------
    config : Configuration, optional
        the configuration to consider for the simulation

    References
    ----------
    https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods

    """

    def __init__(
        self,
        config: Configuration = None,
    ):
        super(RK4St, self).__init__(config)
        self.rng = np.random.default_rng()

    def du_fluct(self, t, u, dt):
        _, scatt_list = get_force_vec(u, self.config, return_list=True)
        dv_tot = np.zeros_like(u[..., :3])
        for scatt in scatt_list:
            # 0 - get scattering rate, laser wavenumber and unit vector for each laser
            rate = scatt["rate"]  # scattering rate
            k = scatt["k"]  # laser wavenumber
            u = scatt["unit_vector"]  # laser unit vector
            Ni = rate * dt  # number of scattered photons
            m = self.config.atom.mass
            # 1 - absorption fluctuation
            # large number of photon approximation
            #   > fluctuation are Gaussian with std = np.sqrt(Ni)
            # note that dN has the same shape as Ni, and can be an array !!
            dN = np.asanyarray(self.rng.normal(loc=0, scale=np.sqrt(Ni)))
            dv_abs = (csts.hbar * k / m) * dN[..., np.newaxis] * u
            dv_tot = dv_tot + dv_abs
            # 2 - emission fluctuation
            # Gaussian approx for random walk, with std = sqrt(Ni/3) for x, y, z
            dNx = np.asanyarray(self.rng.normal(loc=0, scale=np.sqrt(Ni / 3)))
            dNy = np.asanyarray(self.rng.normal(loc=0, scale=np.sqrt(Ni / 3)))
            dNz = np.asanyarray(self.rng.normal(loc=0, scale=np.sqrt(Ni / 3)))
            dN = np.array([dNx.T, dNy.T, dNz.T]).T
            dv_em = (csts.hbar * k / m) * dN
            dv_tot = dv_tot + dv_em

        dx, dy, dz = np.zeros_like(dv_tot.T)
        dvx, dvy, dvz = dv_tot.T
        res = np.array([dx, dy, dz, dvx, dvy, dvz]).T
        return res

    def _iterate(self, t, u, dt):
        """returns the evolution du of u between t and t+dt
        Here we use the fourth order Runge-Kutta method
        """
        # perform step
        # 1) deterministic part
        k1 = self.dudt(t, u)
        k2 = self.dudt(t + 0.5 * dt, u + 0.5 * k1 * dt)
        k3 = self.dudt(t + 0.5 * dt, u + 0.5 * k2 * dt)
        k4 = self.dudt(t + dt, u + k3 * dt)
        du = (dt / 6) * (k1 + 2 * k2 + 2 * k3 + k4)

        # 2) fluctating part
        du_fluct = self.du_fluct(t, u, dt)
        return du + du_fluct


class EulerSt(CustomSimulationBase):
    """A homemade simulator based on simple Euler integration method, taking into
    account spontaneous emission.

    Parameters
    ----------
    config : Configuration, optional
        the configuration to consider for the simulation

    enable_fluct : Boolean, optional (default=True)
        if set to True, fluctations are enabled. Otherwise, the simulator boils
        down to a deterministic Euler simulator

    References
    ----------
    TODO: put here

    """

    def __init__(
        self,
        config: Configuration = None,
        enable_fluct: bool = True,
    ):
        super(EulerSt, self).__init__(config)
        self.enable_fluct = enable_fluct
        self.rng = np.random.default_rng()

    def du_fluct(self, t, u, dt):
        _, scatt_list = get_force_vec(u, self.config, return_list=True)
        dv_tot = np.zeros_like(u[..., :3])
        for scatt in scatt_list:
            # 0 - get scattering rate, laser wavenumber and unit vector for each laser
            rate = scatt["rate"]  # scattering rate
            k = scatt["k"]  # laser wavenumber
            u = scatt["unit_vector"]  # laser unit vector
            Ni = rate * dt  # number of scattered photons
            m = self.config.atom.mass
            # 1 - absorption fluctuation
            # large number of photon approximation
            #   > fluctuation are Gaussian with std = np.sqrt(Ni)
            # note that dN has the same shape as Ni, and can be an array !!
            dN = np.asanyarray(self.rng.normal(loc=0, scale=np.sqrt(Ni)))
            dv_abs = (csts.hbar * k / m) * dN[..., np.newaxis] * u
            dv_tot = dv_tot + dv_abs
            # 2 - emission fluctuation
            # Gaussian approx for random walk, with std = sqrt(Ni/3) for x, y, z
            dNx = np.asanyarray(self.rng.normal(loc=0, scale=np.sqrt(Ni / 3)))
            dNy = np.asanyarray(self.rng.normal(loc=0, scale=np.sqrt(Ni / 3)))
            dNz = np.asanyarray(self.rng.normal(loc=0, scale=np.sqrt(Ni / 3)))
            dN = np.array([dNx.T, dNy.T, dNz.T]).T
            dv_em = (csts.hbar * k / m) * dN
            dv_tot = dv_tot + dv_em

        dx, dy, dz = np.zeros_like(dv_tot.T)
        dvx, dvy, dvz = dv_tot.T
        res = np.array([dx, dy, dz, dvx, dvy, dvz]).T
        return res

    def _iterate(self, t, u, dt):
        """returns the evolution du of u between t and t+dt
        Here we use the fourth order Runge-Kutta method
        """
        # perform step
        # 1) deterministic part
        F = self.get_force(u)
        _, _, _, vx, vy, vz = u.T
        dvx, dvy, dvz = F.T / self.config.atom.mass * dt
        dx = vx * dt + 0.5 * dvx * dt
        dy = vy * dt + 0.5 * dvy * dt
        dz = vz * dt + 0.5 * dvz * dt
        du = np.array([dx, dy, dz, dvx, dvy, dvz]).T

        # 2) fluctating part
        if self.enable_fluct:
            du += self.du_fluct(t, u, dt)
        return du



"""The ``atomsmltr.atoms`` subpackage provides classes to handle atomic species and their transitions.

Examples
---------

Using an existing atom

.. code-block:: python

    from atomsmltr.atoms import Ytterbium  # import `Ytterbium` class
    yb = Ytterbium()  # init an object
    yb.print_info()  # print atom informations

One can also define a new atom from scratch

.. code-block:: python

    from atomsmltr.atoms import Atom
    from atomsmltr.atoms.transitions import DummyTransition
    from scipy import constants as csts
    atom = Atom(mass=4 * csts.m_u, name="Helium")
    transition = DummyTransition("transition", Gamma=1, wavelength=1)
    atom.add_transition(transition, tag="transition")

"""

__all__ = [
    "Atom",
    "Ytterbium",
    "Strontium",
    "Rubidium",
    "DummyTransition",
    "J0J1Transition",
]

from .generic import Atom
from .collection import Ytterbium, Strontium, Rubidium
from .transitions import DummyTransition, J0J1Transition



"""
volumes - 1D zones
=======================

Here we implement 3D zones, namely ``Box``

.. code-block:: python

    from atomsmltr.environment.zones import Box

    pos_box = Box(
        xmin=-10,
        xmax=10,
        ymin=0,
        ymax=5,
        zmin=-8,
        zmax=100,
        target="position",
        action="tag",
        tag="position box",
    )
"""

# % IMPORTS
import numpy as np

# % LOCAL IMPORTS
from .generic import Zone
from ...utils.infostring import InfoString
from ...utils.misc import check_positive_float

# % CLASSES


class Box(Zone):
    """A 3D box with cartesian coordinates

    Parameters
    ----------
    xmin : float
        minimum value for x
    xmax : float
        maximum value for x
    ymin : float
        minimum value for y
    ymax : float
        maximum value for y
    zmin : float
        minimum value for z
    zmax : float
        maximum value for z
    target : str, optional
        the target for the zone, can be "position" or "speed", by default "position"
    action : str, optional
        the action associated to the zone.
        Currently only "stop" is implemented, by default "stop"
    tag : str, optional
        the zone tag
    in_tag : str, optional
        tag for an object inside the zone, by default None
    out_tag : str, optional
        tag for an object inside the zone, by default None


    Example
    -------

    .. code-block:: python

        from atomsmltr.environment.zones import Box

        pos_box = Box(
            xmin=-10,
            xmax=10,
            ymin=0,
            ymax=5,
            zmin=-8,
            zmax=100,
            target="position",
            action="tag",
            tag="position box",
        )

    """

    def __init__(
        self,
        xmin: float,
        xmax: float,
        ymin: float,
        ymax: float,
        zmin: float,
        zmax: float,
        target: str = "position",
        action: str = "stop",
        tag: str = None,
        in_tag: str = None,
        out_tag: str = None,
    ):

        super(Box, self).__init__(
            target=target, action=action, tag=tag, in_tag=in_tag, out_tag=out_tag
        )
        self.xmin = xmin
        self.xmax = xmax
        self.ymin = ymin
        self.ymax = ymax
        self.zmin = zmin
        self.zmax = zmax

    # -- GETTERS & SETTERS

    @property
    def type(self):
        return "3D Box"

    @property
    def xmin(self):
        """float: the minimum value for x"""
        return self.__xmin

    @xmin.setter
    def xmin(self, xmin):
        self.__xmin = float(xmin)

    @property
    def xmax(self):
        """float: the maximum value for x"""
        return self.__xmax

    @xmax.setter
    def xmax(self, xmax):
        self.__xmax = float(xmax)

    @property
    def ymin(self):
        """float: the minimum value for y"""
        return self.__ymin

    @ymin.setter
    def ymin(self, ymin):
        self.__ymin = float(ymin)

    @property
    def ymax(self):
        """float: the maximum value for y"""
        return self.__ymax

    @ymax.setter
    def ymax(self, ymax):
        self.__ymax = float(ymax)

    @property
    def zmin(self):
        """float: the minimum value for z"""
        return self.__zmin

    @zmin.setter
    def zmin(self, zmin):
        self.__zmin = float(zmin)

    @property
    def zmax(self):
        """float: the maximum value for z"""
        return self.__zmax

    @zmax.setter
    def zmax(self, zmax):
        self.__zmax = float(zmax)

    # -- ZONE

    def _in_zone(self, vector):
        x, y, z = vector.T
        in_zone = (
            (x >= self.xmin)
            & (x <= self.xmax)
            & (y >= self.ymin)
            & (y <= self.ymax)
            & (z >= self.zmin)
            & (z <= self.zmax)
        )
        return in_zone.T

    # -- INFOSTRING

    def gen_infostring_obj(self):
        """Generates an info string object"""
        title = self.type
        title = title[:1].upper() + title[1:]  # capitalize first letter
        info = InfoString(title=title)
        info.add_section("Parameters")
        info.add_element("type", "3D Box")
        info.add_element("tag", self.tag)
        info.add_element("in_tag", self.in_tag)
        info.add_element("out_tag", self.tag)
        info.add_element("target", self.target)
        info.add_element("action", self.action)
        info.add_element(f"xmin, xmax", f"{self.xmin, self.xmax}")
        info.add_element(f"ymin, ymax", f"{self.ymin, self.ymax}")
        info.add_element(f"zmin, zmax", f"{self.zmin, self.zmax}")
        info.add_element(f"inverted", f"{self.inverted}")
        return info

    # -- PLOT

    def plot1D(self, ax=None):
        pass

    def plot2D(self, ax=None):
        pass

    def plot3D(self, ax=None):
        pass


class Cylinder(Zone):
    """A cylinder zone

    Parameters
    ----------
    origin : array, shape (3), optional
        the 'center' of the cylinder, i.e. a point on its axis, by default (0, 0, 0)
    direction : array, shape (3), optional
        a vector along the axis of the cylinder, by default (1, 0, 0)
    radius : float, optional
        the cylinder radius, in m or m/s, by default 1.0
    target : str, optional
        the target for the zone, can be "position" or "speed", by default "position"
    action : str, optional
        the action associated to the zone.
        Currently only "stop" is implemented, by default "stop"
    tag : str, optional
        the zone tag
    in_tag : str, optional
        tag for an object inside the zone, by default None
    out_tag : str, optional
        tag for an object inside the zone, by default None
    """

    def __init__(
        self,
        origin: np.ndarray = (0, 0, 0),
        direction: np.ndarray = (1, 0, 0),
        radius: float = 1.0,
        target="position",
        action="stop",
        tag=None,
        in_tag: str = None,
        out_tag: str = None,
    ):

        super(Cylinder, self).__init__(
            target=target, action=action, tag=tag, in_tag=in_tag, out_tag=out_tag
        )
        self.origin = origin
        self.direction = direction
        self.radius = radius

    # -- PROPERTIES

    @property
    def type(self):
        return "3D Cylinder"

    # - origin
    @property
    def origin(self) -> np.ndarray:
        """array, shape (3): cylinder 'center'"""
        return self.__origin

    @origin.setter
    def origin(self, value: np.ndarray):
        value = np.asanyarray(value)
        if value.size != 3:
            raise ValueError("'origin' should be an array of size 3")
        self.__origin = value

    # - direction
    @property
    def direction(self) -> np.ndarray:
        """array, shape (3): cylinder axis direction"""
        return self.__direction

    @direction.setter
    def direction(self, value: np.ndarray):
        value = np.asanyarray(value)
        if value.size != 3:
            raise ValueError("'direction' should be an array of size 3")
        if np.linalg.norm(value) == 0:
            raise ValueError("'the norm of 'direction' cannot be zero")
        self.__direction = value / np.linalg.norm(value)

    # - radius
    @property
    def power(self) -> float:
        """float: cylinder radius (m) or (m/s)"""
        return self._power

    @power.setter
    def power(self, value: float) -> None:
        check_positive_float("power", value)
        self._power = float(value)

    # -- METHODS

    def _in_zone(self, vector):
        # -- compute distance to axis
        r = vector - self.origin
        cross_product = np.cross(r, self.direction)
        distance = np.linalg.norm(cross_product, axis=-1)
        # -- cylinder
        return distance <= self.radius

    # -- INFOSTRING

    def gen_infostring_obj(self):
        """Generates an info string object"""
        title = self.type
        title = title[:1].upper() + title[1:]  # capitalize first letter
        info = InfoString(title=title)
        info.add_section("Parameters")
        info.add_element("type", "3D Cylinder")
        info.add_element("tag", self.tag)
        info.add_element("in_tag", self.in_tag)
        info.add_element("out_tag", self.tag)
        info.add_element("target", self.target)
        info.add_element("action", self.action)
        info.add_element(f"direction", f"{self.direction}")
        info.add_element(f"origin", f"{self.origin}")
        info.add_element(f"radius", f"{self.radius}")
        info.add_element(f"inverted", f"{self.inverted}")
        return info

    # -- PLOT

    def plot1D(self, ax=None):
        pass

    def plot2D(self, ax=None):
        pass

    def plot3D(self, ax=None):
        pass



"""
magpylib wrapper
=======================

Defines a class allowing to wrap a magpylib object into a ``MagneticField`` object
compatible with ``atomsmltr``

see: https://magpylib.readthedocs.io

"""

# % IMPORTS
import numpy as np


# % LOCAL IMPORTS
from .magnetic import MagneticField
from ....utils.infostring import InfoString
import magpylib as magpy


class MagpylibWrapper(MagneticField):
    """A wrapper for magpylib objects

    Parameters
    ----------
    magpy_object : magpy_object
        the object to wrap
    tag : str, optional
        the field tag, by default None
    """

    def __init__(self, magpy_object, tag: str = None):
        super(MagpylibWrapper, self).__init__(tag)
        self.magpy_object = magpy_object

    @property
    def type(self):
        return "magpylib object"

    # -- requested methods for Field
    # pylint : disable=method_hidden
    def _field_value_func(self, position):
        """Returns field value at point position

        position should be an array of shape (3,) or (n1,n2,..,3)
        last axis contains coordinates x, y, z

        NB: position is already checked and converted to an array in the
            `Field` class
        """
        # let's call the magpy get_B function
        B = magpy.getB(self.magpy_object, position, squeeze=False)
        # sqeeeeeeze
        value = B.T
        while value.ndim > position.ndim:
            value = np.squeeze(value, axis=-1)
        return value.T

    def gen_infostring_obj(self):
        """Generates an info string object"""
        unit = self.unit
        title = self.type
        title = title[:1].upper() + title[1:]  # capitalize first letter
        info = InfoString(title=title)
        info.add_section("Parameters")
        info.add_element("type", "magpylib object")
        info.add_element("tag", self.tag)
        # TODO can we have more info ?
        return info



"""
generic fields
=======================

This module implements the generic ``Field`` class, and some perfect fields
(Offset, Gradient, Quadrupole). Those are abstract classes, and actual implementations
are gathered in other packages

See also
---------
atomsmltr.environment.fields.magnetic
"""

# % IMPORTS
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.axes import Axes

from abc import abstractmethod

# % LOCAL IMPORTS
from ..envbase import EnvObject
from ...utils.infostring import InfoString

# % ABSTRACT CLASSES


class Field(EnvObject):
    """A generic class to describe Field objects."""

    def __init__(self, *args, **kwargs):
        super(Field, self).__init__(*args, **kwargs)

    @property
    def vector(self):
        return True

    @property
    @abstractmethod
    def unit():
        """str: returns the unit of the field"""
        pass

    def get_value(self, position: np.ndarray) -> np.ndarray:
        """returns the value of the field at a given position.

        Parameters
        ----------
        position : np.ndarray, shape (,3) or (n1, n2, .., 3)
            positions at which the intensity is computed, given as cartesian coordinates
            in the lab frame.

        Returns
        -------
        value : np.ndarray, same shape as position
            returns the (vector) field value

        Notes
        -------
        `position` is an array_like object, with shape (3,) or (n1, n2, .., 3).
        In all cases, the last dimension contains cordinates (x, y, z), in meter and in the lab frame

        The field value is returned as an array with the same shape as `position`.

        >>> field_value = field.get_value(position)
        >>> X, Y, Z = position.T
        >>> Fx, Fy, Fz = field_value.T
        """
        # Check position
        position = self._check_position_array(position)
        # call hidden function that actually does the computation
        return self._field_value_func(position)

    @abstractmethod
    def _field_value_func(self, position):
        """Actual method for field computation ; defined for each subclass"""

    def get_norm(self, position: np.ndarray) -> np.ndarray:
        """Returns the field norm at a given position in the lab frame

        Parameters
        ----------
        position : np.ndarray, shape (,3) or (n1, n2, .., 3)
            positions at which the intensity is computed, given as cartesian coordinates
            in the lab frame.

        Returns
        -------
        norm : np.ndarray, shape (,1) or (n1, n2, .., 1)
            returns the (scalar) field norm
        """
        F = self.get_value(position)
        Fx, Fy, Fz = F.T
        F_norm = np.sqrt(Fx**2 + Fy**2 + Fz**2).T
        return F_norm

    def plot1D(
        self,
        start: np.ndarray,
        stop: np.ndarray,
        Npoints: int = 100,
        component: str = "Bz",
        ax=None,
        show: bool = False,
        space_scale: float = 1.0,
    ):
        """Plots a 1D line cut of the magnetic field using Matplotlib

        Parameters
        ----------
        start : array-like, shape (3,)
            Starting point (x, y, z) of the line along which to sample the field.
        stop : array-like, shape (3,)
            Ending point (x, y, z) of the line along which to sample the field.
        Npoints : int, optional
            Number of points sampled along the line. Defaults to 100.
        component : str, optional
            Field component to plot. Accepted values are "Bx", "By", "Bz" for vector components,
            or "B" for total field magnitude. Defaults to "Bz".
        ax : Matplotlib Axes, optional
            The matplotlib axis on which to plot.
            If None is given, a new figure is created. Defaults to None.
        show : bool, optional
            Whether to show the figure after calling the method. Defaults to False.
        space_scale : float, optional
            Space coordinates will be multiplied by this when plotting. Defaults to 1.

        Returns
        -------
        ax : Matplotlib Axes
            The axis on which the plot was performed.

        Notes
        ------
        The field is sampled along a straight line between `start` and `stop` using
        `Npoints` equally spaced positions. The field component or magnitude is computed
        and plotted as a function of the distance along the line.

        Examples
        ---------
        >>> field.plot1D(start=[0, 0, -10], stop=[0, 0, 10], Npoints=200)
        >>> field.plot1D(start=[0, 0, -5], stop=[5, 0, 0], component="B", show=True)
        >>> field.plot1D(start=[-5, 0, 0], stop=[5, 0, 0], component="Bx", space_scale=1e-3)

        """
        # Create points along the line
        start = np.array(start)
        stop = np.array(stop)
        t = np.linspace(0, 1, Npoints)
        points = start[None, :] + t[:, None] * (stop - start)[None, :]

        # Compute field
        B = self.get_value(points)

        # Choose component
        if component == "Bx":
            values = B[:, 0]
        elif component == "By":
            values = B[:, 1]
        elif component == "Bz":
            values = B[:, 2]
        elif component == "B":
            values = np.linalg.norm(B, axis=1)
        else:
            raise ValueError(
                f"Unknown component '{component}'. Choose 'Bx', 'By', 'Bz', or 'B'."
            )

        # Compute position along the line (scaled if needed)
        distances = np.linspace(0, np.linalg.norm(stop - start), Npoints) * space_scale

        # Plot
        if ax is None:
            fig, ax = plt.subplots()

        ax.plot(distances, values, label=component)
        ax.set_xlabel("Distance along line")
        ax.set_ylabel(f"Magnetic field [{component}]")
        ax.set_title(f"Field along 1D line from {start} to {stop}")
        ax.grid(True)
        ax.legend()

        if show:
            plt.show()

        return ax

    def plot2D(
        self,
        limits: np.ndarray,
        Npoints: np.ndarray,
        cut: float = 0.0,
        ax=None,
        plane: str = "XY",
        cmap=None,
        show: bool = False,
        space_scale: float = 1.0,
    ):
        """Plots a 2D cut of the field, using Matplotlib streamplot()

        Parameters
        ----------
        limits : array, shape (4,)
            an array of size 4, providing (xmin, xmax, ymin, ymax).
        Npoints : int or array of shape (2,)
            number of points for each dimension,
            either a int or an array of two ints (Nx, Ny).
        cut : float, optional
            coordinate of the third axis for the cut. Defaults to 0.
        ax : Matplotlib Axes, optional
            the matplotlib axis on which to plot.
            If None is given a new figure is created.
            Defaults to None.
        plane : str, optional
            the plane for the cut. Accepted values are "XY", "YZ" and "ZX". Defaults to "XY".
        cmap : Matplotlib cmap, optional
            passed to matplotlib streamplot() function
        show : bool, optional
            whether to show the figure after calling the method. Defaults to False.
        space_scale : float, optional
            space coordinates will be multiplied by this when plotting. Defaults to 1.

        Returns
        -------
        ax : Matplotlib Axes
            the axis on which the plot was performed.

        Notes
        ------
        The limits are given via an array of size 4 'limits', providing providing (xmin, xmax, ymin, ymax)
        Number of points are given with 'Npoints', either as an integer (same value for x and y) or an array of size 2
        the coordinate of the cut axis is given by 'cut'

        Examples
        ---------
        >>> field.plot2D(limits=(-5, 5, -4, 4), Npoints=200)
        >>> field.plot2D(limits=(-5, 5, -4, 4), Npoints=200, cut=-5)
        >>> field.plot2D(limits=(-5, 5, -4, 4), Npoints=(200, 100))

        """
        # - process arguments using the Plottable builtin method
        ax, position, X, Y = self._process_2D_plot_args(
            ax=ax,
            plane=plane,
            limits=limits,
            Npoints=Npoints,
            cut=cut,
        )
        # - compute field
        mag_field = self.get_value(position)
        Bx, By, Bz = mag_field.T
        Bx = Bx.T
        By = By.T
        Bz = Bz.T
        # - get relevant part
        match plane.upper():
            case "XY":
                u = Bx
                v = By

            case "YZ":
                u = By
                v = Bz
            case "ZX":
                u = Bz
                v = Bx

        color = np.sqrt(Bx**2 + By**2 + Bz**2)
        # Transpose if needed, since streamplot is quite strict..
        if not np.allclose(X[0], X):
            X = X.T
            Y = Y.T
            u = u.T
            v = v.T
            color = color.T

        # - plot
        ax.streamplot(X * space_scale, Y * space_scale, u, v, color=color, cmap=cmap)
        ax.set_xlabel(plane.upper()[0])
        ax.set_ylabel(plane.upper()[1])

        # - show ?
        if show:
            plt.show()

        return ax

    def plot3D(
        self,
        limits,
        Npoints,
        ax=None,
        color=None,
        name=None,
        show=False,
        scale=1.0,
        normalize=False,
    ):
        """plots a 3D representation of the field, using Matplotlib quiver()

        Parameters
        ----------
        limits : array, shape (,6)
            limits for the plot (xmin, xmax, ymin, ymax, zmin, zmax)
        Npoints : int or array of shape (,3)
            number of points for each dimension,
            either a int or an array of three ints (Nx, Ny, Nz)
        ax : custom Axes3D, optional
            the axis in which to plot.
            If None is given (default value) a new ax is generated
        color : str, optional
            a matplotlib compatible color. Defaults to None.
        name : str, optional
            the name of the field, passed as a label when plotting. Defaults to None.
        show : bool, optional
            whether the show the figure after calling the method. Defaults to False.
        scale : float, optional
            a scale factor for plotting the arrows (defaults to 1)
        normalize : bool, optional
            if set to True, we normalize the magnetic field to have a max value of 1 before plotting.
            Defaults to False

        Returns
        -------
        ax : Matplotlib Axes
            the axis on which the plot was performed.

        """

        # ------------------------- START ARGUMENT CHECKING ----------------
        # - check plot config
        assert ax is None or isinstance(ax, Axes), "'ax' should be a matplotlib axis."
        # - check axis config
        # limits
        assert np.asanyarray(limits).size == 6, "`limits` should be an array of size 6"
        # Npoints
        Npoints = np.asanyarray(Npoints)
        msg = "`Npoints` should be an int or a list of three ints"
        assert Npoints.size in [1, 3], msg
        assert issubclass(Npoints.dtype.type, np.integer), msg
        # ------------------------- STOP ARGUMENT CHECKING ----------------
        # - init ax (if needed)
        ax = self._init_ax(ax, ax3D=True)
        # - generate grid
        xmin, xmax, ymin, ymax, zmin, zmax = limits
        Nx, Ny, Nz = (Npoints, Npoints, Npoints) if Npoints.size == 1 else Npoints
        grid = np.mgrid[
            xmin : xmax : Nx * 1j, ymin : ymax : Ny * 1j, zmin : zmax : Nz * 1j
        ]
        X, Y, Z = grid
        position = grid.T
        # - get magnetic field
        B = self.get_value(position)
        # - normalize ?
        if normalize:
            B = B / np.max(np.abs(B.ravel()))
        B = B * scale
        Bx, By, Bz = B.T
        # - plot
        ax.quiver(
            X,
            Y,
            Z,
            Bx,
            By,
            Bz,
            label=name,
            color=color,
        )
        # - axes names
        ax.set_xlabel("x")
        ax.set_ylabel("y")
        ax.set_zlabel("z")
        # - show
        if show:
            plt.show()
        return ax

    # -- USEFUL CHECK FUNCTIONS
    def _check_real_number(self, value, name):
        if np.asanyarray(value).size > 1:
            raise ValueError(f"'{name}' should be a scalar")
        if not np.isreal(value):
            raise TypeError(f"'{name}' should be a real numbers")
        return value

    def _check_3D_vector(self, value, name, norm=False):
        value = np.asanyarray(value)
        if value.size != 3:
            raise ValueError(f"'{name}' should be an array of size 3")
        if not np.all(np.isreal(value)):
            raise TypeError(f"'{name}' should be an array of real numbers")
        if norm:
            value = value / np.linalg.norm(value)
        return value


# % TOOL CLASSES


class ConstantField(Field):
    """Generates a constant field

    Parameters
    ----------
    field_value : np.ndarray, shape (3,), optional
        Constant field value, by default (0, 0, 0)
    tag : str, optional
        Field tag, by default None
    """

    def __init__(self, field_value: np.ndarray = (0, 0, 0), tag: str = None):
        self.field_value = field_value
        super(Field, self).__init__(tag)

    # -- getters and setters

    @property
    def field_value(self) -> np.ndarray:
        return self.__field_value

    @field_value.setter
    def field_value(self, value: np.ndarray):
        """field_value (array): constant field value"""
        self.__field_value = self._check_3D_vector(value, "value")

    # -- requested methods for Field
    # pylint : disable=method_hidden
    def _field_value_func(self, position):
        """Returns field value at point position

        position should be an array of shape (3,) or (n1,n2,..,3)
        last axis contains coordinates x, y, z

        NB: position is already checked and converted to an array in the
            `Field` class
        """
        # 'position' already has the right size here
        # as it contains 3D vectors (position)
        # so we can generate an homogeneous field quite easily
        value = position * 0.0 + self.__field_value
        return value

    def gen_infostring_obj(self):
        unit = self.unit
        title = self.type
        title = title[:1].upper() + title[1:]  # capitalize first letter
        info = InfoString(title=title)
        info.add_section("Parameters")
        info.add_element("type", "constant field")
        info.add_element("tag", self.tag)
        info.add_element(f"field_value ({unit})", f"{self.field_value}")
        info.add_element(f"norm ({unit})", f"{np.linalg.norm(self.field_value):.3g}")
        return info


class GradientField(Field):
    """A perfect linear field gradient

    Parameters
    ----------
    origin : array, shape (,3)
        origin for the gradient, in cartesian coordinates in the lab frame
    slope : float
        the slope of the gradient, in 'field unit' / m
    gradient_direction : array, shape (.3)
        the direction of the gradient
    field_direction : array, shae (,3)
        the direction of the field
    offset : float, optional
        an offset for the field amplitude at origin, in 'field unit', by default 0.0
    tag : str, optional
        field tag, by default None

    Notes
    -----
        Note that 'gradient_direction' and 'field_direction' are meant to be
        unit vectors, but the class will take care of normalizing any non normalized entry

    Examples
    ---------

    Create a field pointing along z, with a amplitude increasing linearly along x

    >>> gradient = GradientField(
    ...  origin=(0, 0, 0),
    ...  slope=1,
    ...  gradient_direction=(1, 0, 0),
    ...  field_direction=(0, 0, 1),
    ...  tag="gradient",
    ... )

    """

    def __init__(
        self,
        origin: np.ndarray,
        slope: float,
        gradient_direction: np.ndarray,
        field_direction: np.ndarray,
        offset: float = 0.0,
        tag: str = None,
    ):
        self.slope = slope
        self.offset = offset
        self.origin = origin
        self.gradient_direction = gradient_direction
        self.field_direction = field_direction
        super(GradientField, self).__init__(tag)

    # -- value
    # pylint : disable=method_hidden
    def _field_value_func(self, position):
        """Returns field value at point position

        position should be an array of shape (3,) or (n1,n2,..,3)
        last axis contains coordinates x, y, z

        NB: position is already checked and converted to an array in the
            `Field` class
        """
        # - get X, Y, and Z
        x, y, z = position.T
        x, y, z = x.T, y.T, z.T

        # - get gradient vector angles
        theta = self.__gradient_theta
        phi = self.__gradient_phi

        # - get coordinates w.r.t origin
        x0, y0, z0 = self.origin
        xc = x - x0
        yc = y - y0
        zc = z - z0

        # compute coordinates in rotated frame
        # we want z
        z_rot = (
            xc * np.sin(theta) * np.cos(phi)
            + yc * np.sin(theta) * np.sin(phi)
            + zc * np.cos(theta)
        )

        # - value at position
        value = (self.offset + z_rot * self.slope)[
            ..., np.newaxis
        ] * self.field_direction

        return value

    def gen_infostring_obj(self):
        """Generates an info string object"""
        unit = self.unit
        title = self.type
        title = title[:1].upper() + title[1:]  # capitalize first letter
        info = InfoString(title=title)
        info.add_section("Parameters")
        info.add_element("type", "perfect gradient")
        info.add_element("tag", self.tag)
        info.add_element(f"slope ({unit}/m)", f"{self.slope:.3g}")
        info.add_element("gradient direction", f"{self.gradient_direction}")
        info.add_element("field direction", f"{self.field_direction}")
        info.add_element(f"origin (m)", f"{self.origin}")
        info.add_element(f"offset ({unit})", f"{self.offset:3g}")
        return info

    # -- getters and setters
    # -
    @property
    def slope(self) -> float:
        """float: the field amplitude gradient slope"""
        return self.__slope

    @slope.setter
    def slope(self, value: float):
        self.__slope = self._check_real_number(value, "slope")

    # -
    @property
    def offset(self) -> float:
        """float: the field amplitude value at origin"""
        return self.__offset

    @offset.setter
    def offset(self, value: float):
        self.__offset = self._check_real_number(value, "offset")

    # -
    @property
    def origin(self) -> np.ndarray:
        """array, shape (,3): the origin for the gradient"""
        return self.__origin

    @origin.setter
    def origin(self, value: np.ndarray):
        self.__origin = self._check_3D_vector(value, "origin")

    # -
    @property
    def gradient_direction(self) -> np.ndarray:
        """array, shape (,3): the gradient direction"""
        return self.__gradient_direction

    @gradient_direction.setter
    def gradient_direction(self, value: np.ndarray):
        value = self._check_3D_vector(value, "gradient_direction", norm=True)
        assert np.allclose(
            np.linalg.norm(value), 1
        ), "We did not manage to normalize gradient_direction, something is fishy.."
        # compute angles
        ux, uy, uz = value
        theta = np.arctan2(np.sqrt(ux**2 + uy**2), uz)
        phi = np.arctan2(uy, ux)
        self.__gradient_direction = value
        self.__gradient_theta = theta
        self.__gradient_phi = phi

    # -
    @property
    def field_direction(self) -> np.ndarray:
        """array, shape (,3): the field direction"""
        return self.__field_direction

    @field_direction.setter
    def field_direction(self, value: np.ndarray):
        self.__field_direction = self._check_3D_vector(
            value, "field_direction", norm=True
        )


class BaseQuadrupoleField(Field):
    """A perfect quadrupole field

    Parameters
    ----------
    origin : array, shape (,3)
        origin for the quadrupole, in cartesian coordinates in the lab frame
    strong_axis : array, shape (,3)
        the direction of the strong axis
    slope : float
        the gradient of the weak axes
    tag : str, optional
        field tag, by default None

    Notes
    ------

    Note that 'strong_axis' is meant to be a unit vector, but the class will take care of normalizing
    any non normalized entry

    """

    def __init__(
        self,
        origin: np.ndarray,
        strong_axis: np.ndarray,
        slope: float,
        tag: str = None,
    ):
        self.slope = slope
        self.origin = origin
        self.strong_axis = strong_axis

        super(BaseQuadrupoleField, self).__init__(tag)

    # -- getters and setters
    # -
    @property
    def slope(self) -> float:
        """float: the gradient of the weak axes"""
        return self.__slope

    @slope.setter
    def slope(self, value: float):
        self.__slope = self._check_real_number(value, "slope")

    # -
    @property
    def origin(self) -> np.ndarray:
        """array, shape (3,): the origin of the quadrupole"""
        return self.__origin

    @origin.setter
    def origin(self, value: np.ndarray):
        self.__origin = self._check_3D_vector(value, "origin")

    # -
    @property
    def strong_axis(self) -> np.ndarray:
        """array, shape (3,): the direction of the strong axis"""
        return self.__strong_axis

    @strong_axis.setter
    def strong_axis(self, value: np.ndarray):
        value = self._check_3D_vector(value, "strong_axis", norm=True)
        assert np.allclose(
            np.linalg.norm(value), 1
        ), "We did not manage to normalize strong_axis, something is fishy.."
        # compute angles
        ux, uy, uz = value
        theta = np.arctan2(np.sqrt(ux**2 + uy**2), uz)
        phi = np.arctan2(uy, ux)
        self.__strong_axis = value
        self.__theta = theta
        self.__phi = phi

    def gen_infostring_obj(self):
        """Generates an info string object"""
        unit = self.unit
        title = self.type
        title = title[:1].upper() + title[1:]  # capitalize first letter
        info = InfoString(title=title)
        info.add_section("Parameters")
        info.add_element("type", "perfect quadrupole")
        info.add_element("tag", self.tag)
        info.add_element(f"slope ({unit}/m)", f"{self.slope:.3g}")
        info.add_element("strong axis", f"{self.strong_axis}")
        info.add_element(f"origin (m)", f"{self.origin}")
        return info


class QuadrupoleFieldX(BaseQuadrupoleField):
    """A perfect quadrupole field with strong axis along X

    Parameters
    ----------
    origin : array, shape (,3)
        origin for the quadrupole, in cartesian coordinates in the lab frame
    slope : float
        the gradient of the weak axes
    tag : str, optional
        field tag, by default None

    Notes
    ------

    Note that 'strong_axis' is meant to be a unit vector, but the class will take care of normalizing
    any non normalized entry

    """

    def __init__(
        self,
        origin: np.ndarray,
        slope: float,
        tag: str = None,
    ):
        super(QuadrupoleFieldX, self).__init__(
            origin=origin, slope=slope, tag=tag, strong_axis=(1, 0, 0)
        )

    def _field_value_func(self, position):
        """Returns field value at point position

        position should be an array of shape (3,) or (n1,n2,..,3)
        last axis contains coordinates x, y, z

        NB: position is already checked and converted to an array in the
            `Field` class
        """
        # - get X, Y, and Z
        x, y, z = position.T

        # - get coordinates w.r.t origin
        x0, y0, z0 = self.origin
        xc, yc, zc = x - x0, y - y0, z - z0

        # - compute
        slope = self.slope
        value = np.array([-2 * slope * xc, slope * yc, slope * zc]).T

        return value


class QuadrupoleFieldY(BaseQuadrupoleField):
    """A perfect quadrupole field with strong axis along Y

    Parameters
    ----------
    origin : array, shape (,3)
        origin for the quadrupole, in cartesian coordinates in the lab frame
    slope : float
        the gradient of the weak axes
    tag : str, optional
        field tag, by default None

    Notes
    ------

    Note that 'strong_axis' is meant to be a unit vector, but the class will take care of normalizing
    any non normalized entry

    """

    def __init__(
        self,
        origin: np.ndarray,
        slope: float,
        tag: str = None,
    ):
        super(QuadrupoleFieldY, self).__init__(
            origin=origin, slope=slope, tag=tag, strong_axis=(0, 1, 0)
        )

    def _field_value_func(self, position):
        """Returns field value at point position

        position should be an array of shape (3,) or (n1,n2,..,3)
        last axis contains coordinates x, y, z

        NB: position is already checked and converted to an array in the
            `Field` class
        """
        # - get X, Y, and Z
        x, y, z = position.T

        # - get coordinates w.r.t origin
        x0, y0, z0 = self.origin
        xc, yc, zc = x - x0, y - y0, z - z0

        # - compute
        slope = self.slope
        value = np.array([slope * xc, -2 * slope * yc, slope * zc]).T

        return value


class QuadrupoleFieldZ(BaseQuadrupoleField):
    """A perfect quadrupole field with strong axis along Z

    Parameters
    ----------
    origin : array, shape (,3)
        origin for the quadrupole, in cartesian coordinates in the lab frame
    slope : float
        the gradient of the weak axes
    tag : str, optional
        field tag, by default None

    Notes
    ------

    Note that 'strong_axis' is meant to be a unit vector, but the class will take care of normalizing
    any non normalized entry

    """

    def __init__(
        self,
        origin: np.ndarray,
        slope: float,
        tag: str = None,
    ):
        super(QuadrupoleFieldZ, self).__init__(
            origin=origin, slope=slope, tag=tag, strong_axis=(0, 0, 1)
        )

    def _field_value_func(self, position):
        """Returns field value at point position

        position should be an array of shape (3,) or (n1,n2,..,3)
        last axis contains coordinates x, y, z

        NB: position is already checked and converted to an array in the
            `Field` class
        """
        # - get X, Y, and Z
        x, y, z = position.T

        # - get coordinates w.r.t origin
        x0, y0, z0 = self.origin
        xc, yc, zc = x - x0, y - y0, z - z0

        # - compute
        slope = self.slope
        value = np.array([slope * xc, slope * yc, -2 * slope * zc]).T

        return value


class QuadrupoleField(BaseQuadrupoleField):
    """A perfect quadrupole field with strong axis along a given vector

    Parameters
    ----------
    origin : array, shape (,3)
        origin for the quadrupole, in cartesian coordinates in the lab frame
    strong_axis : array, shape (,3)
        the direction of the strong axis
    slope : float
        the gradient of the weak axes
    tag : str, optional
        field tag, by default None

    Notes
    ------

    Note that 'strong_axis' is meant to be a unit vector, but the class will take care of normalizing
    any non normalized entry

    """

    def __init__(
        self,
        origin: np.ndarray,
        strong_axis: np.ndarray,
        slope: float,
        tag: str = None,
    ):
        super(QuadrupoleField, self).__init__(
            origin=origin, slope=slope, tag=tag, strong_axis=strong_axis
        )

    def _field_value_func(self, position):
        """Returns field value at point position

        position should be an array of shape (3,) or (n1,n2,..,3)
        last axis contains coordinates x, y, z

        NB: position is already checked and converted to an array in the
            `Field` class
        """
        # - get X, Y, and Z
        x, y, z = np.moveaxis(position, -1, 0)

        # - get coordinates w.r.t origin
        x0, y0, z0 = self.origin
        xc, yc, zc = x - x0, y - y0, z - z0
        rc = np.stack((xc, yc, zc), axis=-1)

        # - compute
        slope = self.slope
        s = self.strong_axis / np.linalg.norm(self.strong_axis)
        projection = np.sum(rc * s, axis=-1, keepdims=True)

        value = slope * projection * s - (slope / 2) * (rc - projection * s)

        return value



"""The ``atomsmltr.environment.lasers``\ subpackage provides definitions for lasers

It defines two main classes :

+ ``LaserBeam``    > to manage laser intensity, propagation, etc.
+ ``Polarization`` > to specifically manage the polarization

Examples
---------

Setup a Gaussian beam

.. code-block:: python

    from atomsmltr.environment.lasers import GaussianLaserBeam
    from atomsmltr.environment.lasers.polarization import CircularLeft


    beam = GaussianLaserBeam(
        wavelength=399e-9,
        waist=50e-6,
        power=30e-3,
        waist_position=(0, 0, 0),
        direction=(0, 0, 1),
        polarization=CircularLeft(),
    )

See also
--------
atomsmltr.environment.lasers.beams
atomsmltr.environment.lasers.polarization

"""

__all__ = [
    "GaussianLaserBeam",
    "PlaneWaveLaserBeam",
    "LaserBeam",
    "Polarization",
    "Vertical",
    "Horizontal",
    "CircularLeft",
    "CircularRight",
    "Vector",
    "Linear",
]

from .beams import GaussianLaserBeam, PlaneWaveLaserBeam, LaserBeam
from .polarization import (
    Polarization,
    Vertical,
    Horizontal,
    CircularLeft,
    CircularRight,
    Vector,
    Linear,
)



"""
atomSmltr
================

a python package for atomic physics simulations.

Package structure
------------------

| ``atomsmltr.atoms``        : atomic species handling (physical properties, electronic transitions)
| ``atomsmltr.datapipe``     : manage data generated by the simulations (save, load..)
| ``atomsmltr.environment``  : define the atom environment (lasers, mag. fields, zones)
| ``atomsmltr.simulation``   : manage simulation configuration & run simulations
| ``atomsmltr.utils``        : misc tools
| ``atomsmltr.examples``     : contain pre-made example configurations

"""

__version__ = "0.0.1"
__author__ = "Alexandre Dareau"
__all__ = [
    "atoms",
    "datapipe",
    "lasers",
    "zones",
    "fields",
    "simulation",
    "utils",
    "examples",
]

from atomsmltr import atoms
from atomsmltr import datapipe
from atomsmltr.environment import lasers, zones, fields
from atomsmltr import simulation
from atomsmltr import utils
from atomsmltr import examples
